"""The Harbor sub-package exposes the core integrations of the Harbor container image registry connected to AdaLab."""

from .harbor import (
    archive_image,
    create_image_metadata,
    edit_image_metadata,
    get_all_metadata,
    get_image_digest,
    get_image_id,
    get_image_metadata,
    get_image_metadata_id,
    get_project_repositories,
    get_project_stats,
    get_projects,
    get_publish_logs,
    get_repositories_by_author,
    get_repositories_by_type,
    get_repository_tags,
    get_types,
    request_harbor,
    restore_image,
    update_image_state,
)

__all__ = [
    "archive_image",
    "create_image_metadata",
    "edit_image_metadata",
    "get_all_metadata",
    "get_image_digest",
    "get_image_id",
    "get_image_metadata",
    "get_image_metadata_id",
    "get_project_repositories",
    "get_projects",
    "get_project_stats",
    "get_publish_logs",
    "get_repositories_by_author",
    "get_repositories_by_type",
    "get_repository_tags",
    "get_types",
    "request_harbor",
    "restore_image",
    "update_image_state",
]

__title__ = "adalib Harbor"



"""Adalab Volumes Module"""

from .volumes import (
    create_volume,
    delete_volume,
    get_all_volume_mounts_for_app,
    get_all_volume_mounts_for_user,
    get_all_volumes,
    get_volume,
    mount_volume_to_user_lab,
    unmount_volume_from_user_lab,
    update_mounted_volume_to_user_lab,
    update_volume,
)

__all__ = [
    "get_all_volumes",
    "get_volume",
    "get_all_volume_mounts_for_app",
    "get_all_volume_mounts_for_user",
    "create_volume",
    "mount_volume_to_user_lab",
    "update_volume",
    "update_mounted_volume_to_user_lab",
    "delete_volume",
    "unmount_volume_from_user_lab",
]
__title__ = "adalib Volumes"



"""The Lab sub-package exposes the core integrations of the user's Lab environment in AdaLab."""

from .lab import (
    build_image_from_git,
    build_image_from_lab,
    convert_token_to_value,
    delete_files,
    download_file,
    get_available_kernels,
    get_build_logs,
    get_build_status,
    get_config_options,
    get_installed_kernels,
    get_kernel_metadata_id,
    get_lab_files,
    get_lab_images,
    get_lab_logs,
    get_lab_status,
    install_kernel,
    move_file,
    stop_lab,
    uninstall_kernel,
    who_am_i,
)

__all__ = [
    "build_image_from_git",
    "build_image_from_lab",
    "convert_token_to_value",
    "download_file",
    "delete_files",
    "get_available_kernels",
    "get_build_logs",
    "get_build_status",
    "get_config_options",
    "get_installed_kernels",
    "get_kernel_metadata_id",
    "get_lab_files",
    "get_lab_images",
    "get_lab_logs",
    "get_lab_status",
    "install_kernel",
    "move_file",
    "stop_lab",
    "uninstall_kernel",
    "who_am_i",
]
__title__ = "adalib Lab"



"""The images sub-package exposes the core integrations of the image database connected to AdaLab."""

from .pictures import get_picture, post_picture, post_picture_url

__all__ = [
    "get_picture",
    "post_picture",
    "post_picture_url",
]

__tile__ = "adalib Images"



import logging
import urllib.parse
from json import dumps

import pandas as pd
import requests_oauthlib
import supersetapiclient.base
import supersetapiclient.charts
import supersetapiclient.client
import supersetapiclient.dashboards
import supersetapiclient.databases
import supersetapiclient.datasets
import supersetapiclient.exceptions
import supersetapiclient.saved_queries
from adalib_auth import config, keycloak

from ..exceptions import QueryLimitReached

log = logging.getLogger(__name__)


class RunWrapper:
    """Wrapper for converting results from Superset into pandas dataframes
    seamlessly.
    """

    def __init__(self, data, columns):
        """Init the RunWrapper

        :param data: data to be converted to a dataframe
        :type data: dict
        :param columns: columns describing the types of the dataframe
        :type columns: list
        """
        self._data = data
        self._columns = columns

    def __iter__(self):
        """__iter__ implementation for the RunWrapper

        :return: iterator of the data and columns
        :rtype: iter
        """
        return iter((self._data, self._columns))

    def as_df(self):
        """Converts an SQL query result as returned by Superset to a dataframe.

        :return: dataframe constructed from the data + columns
        :rtype: pandas.DataFrame
        """
        #
        dtypes_map = {
            "STRING": "str",
        }
        colspec = {
            c["name"]: dtypes_map[c["type"]]
            for c in self._columns
            if c["type"] in dtypes_map
        }
        df = pd.DataFrame.from_dict(self._data)
        return df.astype(colspec, errors="ignore")


class ListWrapper:
    """Wrapper for returning query results as a list."""

    def __init__(self, objs, df_cols):
        self._objs = objs
        self._df_cols = df_cols

    def __iter__(self):
        """iterator for the ListWrapper objects.

        :return: an iterator for objects in the class instantiation
        :rtype: list_iterator
        """
        return iter(self._objs)

    def __len__(self):
        """len implementation for the ListWrapper.

        :return: lenght of the list of objects in the class instantiation
        :rtype: int
        """
        return len(self._objs)

    def as_df(self):
        """Converts the ListWrapper input data to a pandas.DataFrame.

        :return: dataframe constructed from the data + columns
        :rtype: pandas.DataFrame
        """
        data = [
            {k: getattr(o, f) for f, k in self._df_cols.items()}
            for o in self._objs
        ]
        df = pd.DataFrame(data=data)
        if len(df) and "id" in self._df_cols:
            df.set_index("id", inplace=True)
        return df


class DbListWrapper(ListWrapper):
    """Wrapper for the ListWrapper for handling databases."""

    def __init__(self, databases, *args, **kwargs):
        self._databases = databases
        super().__init__(*args, **kwargs)

    def as_df(self):
        """Add database name to objects returned as a pandas.DataFrame.

        :return: dataframe constructed from the data + columns
        :rtype: pandas.DataFrame
        """
        db_map = {d.id: d.database_name for d in self._databases}
        db_id_field = {v: k for k, v in self._df_cols.items()}["database_id"]
        for o in self._objs:
            o.database_name = db_map.get(getattr(o, db_id_field))
        return super().as_df()


class ListMixIn:
    """MixIn for converting to a list."""

    DATAFRAME_COLS = {}

    def all(self):
        """returns everything that can be found.

        :return: a list of all objects found
        :rtype: list
        """
        return self.find()

    def find(self, *args, **kwargs):
        """not implemented."""
        raise NotImplementedError()

    def fetch(self, *args, **kwargs):
        """Wrapper for the supersetapiclient.get method returning a SavedQueries/Dashboards/Charts/Datasets/Databases object corresponding to the given id.

        :param id: valid object id
        :type id: int (or castable to int)
        :return: Superset object
        :rtype: supersetapiclient.().SavedQueries/Dashboards/Charts/Datasets/Databases
        """
        try:
            kwargs["id"] = int(kwargs["id"])
            res = self.get(*args, **kwargs)
            return res
        except ValueError:
            raise ValueError("'id' must be of type int, or castable to int.")


class SavedQueries(supersetapiclient.saved_queries.SavedQueries, ListMixIn):
    """Wrapper for saved queries in SuperSet.

    :param supersetapiclient: SavedQueries from the superset api client
    :type supersetapiclient: supersetapiclient.saved_queries.SavedQueries
    :param ListMixIn: mixin for returning superset results as lists
    :type ListMixIn: ListMixIn
    """

    DATAFRAME_COLS = {
        "id": "id",
        "label": "name",
        "description": "description",
        "db_id": "database_id",
        "database_name": "database_name",
        "sql": "sql",
    }

    def find(self, *args, **kwargs):
        """wrapper for the supersetapiclient.find method returning a pandas
        DataFrame object with the found saved queries

        :return: A list wrapper for database objects returned
        :rtype: DbListWrapper
        """
        res = super().find(*args, **kwargs)
        dbs = self.client.databases.all()
        return DbListWrapper(
            objs=res, databases=dbs, df_cols=self.DATAFRAME_COLS
        )


class Dashboards(supersetapiclient.dashboards.Dashboards, ListMixIn):
    """Wrapper for the supersetapiclient for returning dashboards from Superset."""

    DATAFRAME_COLS = {
        "id": "id",
        "dashboard_title": "name",
        "published": "published",
        "changed_on": "changed_on",
        "changed_by_name": "changed_by_name",
    }

    def find(self, *args, **kwargs):
        """wrapper for the supersetapiclient.find method returning a pandas
        DataFrame object with the found dashboards

        :return: list of dashboards
        :rtype: list
        """
        res = super().find(*args, **kwargs)
        return ListWrapper(objs=res, df_cols=self.DATAFRAME_COLS)


class Chart(supersetapiclient.charts.Chart):
    """Wrapper for the supersetapiclient chart."""

    def iframe(self, width: int = 600, height: int = 400) -> str:
        """Returns a chart as an iframe to be rendered in a notebook.

        :param width: width of the iframe
        :type width: int
        :param height: height of the iframe
        :type height: int

        :return: an iframe containing the chart to be rendered
        :rtype: str
        """
        base_url = self._parent.client.external_host
        form_data = urllib.parse.quote(dumps(self.params))
        return f"""\
<iframe
  width="{width}"
  height="{height}"
  seamless
  frameBorder="0"
  scrolling="no"
  src="{base_url}/superset/explore/?form_data={form_data}&standalone=1&height={height}"
>
</iframe>
"""

    def embed(self, width: int = 600, height: int = 400) -> None:
        """when called from a notebook, displays the chart in an iframe.

        :param width: width of the iframe, defaults to 600
        :type width: int, optional
        :param height: height of the iframe, defaults to 400
        :type height: int, optional
        """
        from IPython.display import HTML, display

        display(HTML(self.iframe(width=width, height=height)))


class Charts(supersetapiclient.charts.Charts, ListMixIn):
    """Wrapper for the supersetapiclient charts."""

    base_object = Chart
    DATAFRAME_COLS = {
        "id": "id",
        "slice_name": "name",
        "description": "description",
        "datasource_type": "datasource_type",
        "datasource_id": "datasource_id",
        "viz_type": "viz_type",
    }

    def find(self, *args, **kwargs):
        """wrapper for the supersetapiclient.find method returning a pandas
        DataFrame object with the found charts

        :return: list of charts
        :rtype: list
        """
        res = super().find(*args, **kwargs)
        return ListWrapper(objs=res, df_cols=self.DATAFRAME_COLS)


class Datasets(supersetapiclient.datasets.Datasets, ListMixIn):
    """Wrapper for the supersetapiclient datasets."""

    DATAFRAME_COLS = {
        "id": "id",
        "table_name": "name",
        "description": "description",
        "kind": "kind",
        "sql": "sql",
        "database_id": "database_id",
        "database_name": "database_name",
        "schema": "schema",
    }

    def find(self, *args, **kwargs):
        """wrapper for the supersetapiclient.find method returning a pandas
        DataFrame object with the found datasets

        :return: list of datasets
        :rtype: list
        """
        res = super().find(*args, **kwargs)
        dbs = self.client.databases.all()
        return DbListWrapper(
            objs=res, databases=dbs, df_cols=self.DATAFRAME_COLS
        )


class Databases(supersetapiclient.databases.Databases, ListMixIn):
    """Wrapper for the supersetapiclient databases."""

    DATAFRAME_COLS = {
        "id": "id",
        "database_name": "name",
    }

    def find(self, *args, **kwargs):
        """wrapper for the supersetapiclient.find method returning a pandas
        DataFrame object with the found databases

        :return: list of databases
        :rtype: list
        """
        res = super().find(*args, **kwargs)
        return ListWrapper(objs=res, df_cols=self.DATAFRAME_COLS)


class OAuthSupersetClient(supersetapiclient.client.SupersetClient):
    """Class to handle OAuth2 login to Superset for a programmatic client."""

    dashboards_cls = Dashboards
    charts_cls = Charts
    datasets_cls = Datasets
    databases_cls = Databases
    saved_queries_cls = SavedQueries

    def __init__(self, *args, **kwargs):
        """requires and OAuth token valid for authenticating with the Superset API."""
        adalib_config = config.get_config()
        self.oauth_token = kwargs.pop("oauth_token", None)

        self.external_host = adalib_config.SERVICES["superset"]["external"]
        self.internal_host = adalib_config.SERVICES["superset"]["url"]
        kwargs["host"] = (
            self.external_host
            if adalib_config.ENVIRONMENT == "external"
            else self.internal_host
        )
        super().__init__(*args, **kwargs)

    def authenticate(self):
        """
        Authenticate with OAuth using our KeyCloak setup. We want to defer authentication to this point, to allow
        creating a client instance without triggering any requests.

        If no token is given, and we're in a JupyterHub context where we have a JupyterHub API token, then we can
        translate that token to a Superset token using Keycloak token exchange.

        :return: authentication response as json
        :rtype: json
        """
        adalib_config = config.get_config()
        if not self.oauth_token:
            self.oauth_token = keycloak.get_client_token(
                audience_client_id=adalib_config.KEYCLOAK_CLIENTS["superset"]
            )
        session = requests_oauthlib.OAuth2Session(token=self.oauth_token)
        response = session.post(
            self.login_endpoint,
            json={
                "provider": "oauth",
                "refresh": "true",
            },
        )
        response.raise_for_status()
        return response.json()

    def run(self, *args, **kwargs):
        """Runs a query against the Superset API

        :raises QueryLimitReached: if the query limit is reached against the API
        :return: a pandas.DataFrame with the results of the query
        :rtype: RunWrapper
        """
        try:
            columns, data = super().run(*args, **kwargs)
        except supersetapiclient.exceptions.QueryLimitReached as e:
            raise QueryLimitReached(*e.args)
        return RunWrapper(columns=columns, data=data)

    @property
    def _sql_endpoint(self) -> str:
        """returns the sql query endpoint from the Superset API

        :return: url endpoint for executing sql queries
        :rtype: str
        """
        return self.join_urls(self.base_url, "execute_sql_json/")



import logging

import requests
from adalib_auth import config, keycloak


def build_request_url(path: str) -> str:
    """Build the URL to hit adaboard on.

    :param path: The relative path to hit adaboard on.
    :type path: str
    :return: The complete URL to hit adaboard on.
    :rtype: str
    """
    adalib_config = config.get_config()
    url_scheme = (
        "https://" if adalib_config.ENVIRONMENT == "external" else "http://"
    )
    api_extension = (
        "/adaboard/api" if adalib_config.ENVIRONMENT == "external" else ""
    )
    return (
        url_scheme
        + adalib_config.SERVICES["adaboard-api"]["netloc"]
        + api_extension
        + f"/{path}"
    )


def get_all_pages(api_url: str, params: dict = None) -> list[dict]:
    """
    Fetch all pages of data from a paginated API and return the combined results.

    :param api_url: The base API URL to request data from.
    :type api_url: str
    :param params: Additional query parameters for the API request.
    :type params: dict
    :return: A list of items from all pages.
    :rtype: list[dict]
    """
    all_responses = []
    current_page = 1

    while True:
        if params is None:
            params = {}
        params["page"] = current_page
        params["size"] = 100

        response = request_adaboard(
            path=api_url, method=requests.get, params=params
        ).json()

        all_responses.extend(response.get("items"))
        if response.get("page") >= response.get("pages"):
            break

        current_page += 1

    return all_responses


def get_user(
    include_notification_preferences: bool = False,
) -> dict:
    """Get user data.

    :param user_id: User ID
    :type user_id: str
    :param include_notification_preferences: Flag whether the notification
        preferences should be included
    :type include_notification_preferences: bool
    """
    user = request_adaboard(path="users/self").json()
    notifications = request_adaboard(path="nosy/settings/").json()
    user["notifications"] = notifications

    if include_notification_preferences:
        notification_preferences = request_adaboard(
            path="nosy/preferences/"
        ).json()
        notification_preferences = [
            _unravel_notification(x)
            for x in notification_preferences["preferences"]
        ]
        user["notifications"]["preferences"] = notification_preferences

    return user


def request_adaboard(
    path: str,
    method=requests.get,
    omit_content_type: bool = False,
    **kwargs,
) -> requests.models.Response:
    """Function to hit the adaboard api at the specified url and raise for error on response.

    Note: All kwargs passed are directly sent through to the request function.

    :param path: path to hit adaboard on (base url will be injected automatically)
    :type path: str
    :param method: method to use to query adaboard, defaults to requests.get
    :type method: function, optional
    :return: response from adaboard
    :rtype: response
    """
    # Construct authentication header for adaboard
    adalib_config = config.get_config()
    adaboard_token = keycloak.get_client_token(
        audience_client_id=adalib_config.KEYCLOAK_CLIENTS["adaboard-api"]
    )

    headers = {
        "authorization": f"Bearer {adaboard_token['access_token']}",
        "Accept": "application/json",
    }
    if not omit_content_type:
        headers["Content-Type"] = "application/json"

    request_url = build_request_url(path)

    # Query adaboard and raise for status
    try:
        response = method(request_url, headers=headers, **kwargs)
        response.raise_for_status()
        return response
    except requests.exceptions.HTTPError as e:
        try:
            error_message = e.response.json()["detail"]
        except ValueError:
            error_message = e.response.text

        logging.error(
            f"""HTTP error occurred: {e} - """
            f"""Error details: {error_message}"""
        )
        raise


def _unravel_notification(notif: dict) -> dict:
    """Helper for parsing the notification preferences."""
    data = notif["name"].split("__")
    for d in data:
        key, value = d.split("=")
        notif[key.lower()] = value
    del notif["name"]
    return notif



"""The Adaboard sub-package exposes the core integrations with the Adaboard API."""

from .adaboard import (
    _unravel_notification,
    build_request_url,
    get_all_pages,
    get_user,
    request_adaboard,
)

__all__ = [
    "_unravel_notification",
    "build_request_url",
    "get_all_pages",
    "get_user",
    "request_adaboard",
]
__title__ = "adalib Adaboard"



"""Utility functions for working with volumes."""


def validate_volume_mount(volume_mount: dict[str,]):
    """Validate a volume mount object.

    :param volume_mount: The volume mount object to validate
    :type volume_mount: dict[str,]
    """
    assert "volume_id" in volume_mount, "Volume ID is required"
    assert "mount_path" in volume_mount, "Mount path is required"
    assert "read_only" in volume_mount, "Read-only flag is required"
    assert isinstance(
        volume_mount["volume_id"], int
    ), "Volume ID must be an integer"
    assert isinstance(
        volume_mount["mount_path"], str
    ), "Mount path must be a string"
    assert isinstance(
        volume_mount["read_only"], bool
    ), "Read-only flag must be a boolean"



"""AdaLab Shared Volumes Module."""

from typing import Optional

import requests
from requests import Response

from adalib.adaboard import request_adaboard

from ..utils import validate_acl


def get_all_volumes(
    include_all: bool = False, allowed_to_mount: bool = True
) -> list[dict[str,]]:
    """Get all AdaLab Shared Volumes currently available.

    :param include_all: If True, return all volumes. If False, return only volumes
        that the user is allowed to view.
    :type include_all: bool
    :param allowed_to_mount: If True, return only volumes that the user is allowed to
        mount. If False, return all volumes.
    :type allowed_to_mount: bool

    :return: List of AdaLab Shared Volumes.
    :rtype: list[dict[str,]]
    """
    params = {"can_mount": allowed_to_mount}
    if not include_all:
        user_id = request_adaboard(path="users/self").json()["user_id"]
        params["user_id"] = user_id

    return request_adaboard("volumes", params=params).json()


def get_volume(volume_id: int) -> dict[str,]:
    """Get a specific AdaLab Shared Volume.

    :param volume_id: The ID of the volume to get.
    :type volume_id: int
    :return: The AdaLab Shared Volume.
    :rtype: dict[str,]
    """
    return request_adaboard(f"volumes/{volume_id}").json()


def get_all_volume_mounts_for_app(app_id: str) -> list[dict[str,]]:
    """Get all AdaLab Shared Volumes mounted to a specific AdaLab App.

    :param app_id: The ID of the AdaLab App.
    :type app_id: str
    :return: List of AdaLab Shared Volume mounts to the AdaLab App.
    :rtype: list[dict[str,]]
    """
    return request_adaboard(
        "volumes/mounts/app", params={"app_id": app_id}
    ).json()


def get_all_volume_mounts_for_user(
    user_id: Optional[str] = None,
) -> list[dict[str,]]:
    """Get all AdaLab Shared Volumes mounted to a specific AdaLab User's Lab.

    :param user_id: The ID of the AdaLab User.
    :type user_id: Optional[str]
    :return: List of AdaLab Shared Volume mounts to the AdaLab User's Lab.
    :rtype: list[dict[str,]]
    """
    params = dict()
    if user_id is not None:
        params["user_id"] = user_id

    return request_adaboard("volumes/mounts/user", params=params).json()


def create_volume(
    name: str,
    description: str,
    size: int,
    view_acl_type: str = "public",
    view_acl_list: list[str] = [],
    mount_acl_type: str = "public",
    mount_acl_list: list[str] = [],
    edit_acl_type: str = "public",
    edit_acl_list: list[str] = [],
) -> int:
    """Create a new AdaLab Shared Volume.

    :param name: The name of the volume.
    :type name: str
    :param description: The description of the volume.
    :type description: str
    :param size: The size of the volume in GB.
    :type size: int
    :param view_acl_type: The ACL type for viewing the volume. Default is "public".
    :type view_acl_type: str
    :param view_acl_list: The list of users or groups that can view the volume.
    :type view_acl_list: list[str]
    :param mount_acl_type: The ACL type for mounting the volume. Default is "public".
    :type mount_acl_type: str
    :param mount_acl_list: The list of users or groups that can mount the volume.
    :type mount_acl_list: list[str]
    :param edit_acl_type: The ACL type for editing the volume. Default is "public".
    :type edit_acl_type: str
    :param edit_acl_list: The list of users or groups that can edit the volume.
    :type edit_acl_list: list[str]

    :return: The ID of the created volume.
    :rtype: int
    """
    validate_acl("volume", False, view_acl_type, mount_acl_type, edit_acl_type)
    assert size >= 1, "Volume size must be at least 1GB"

    acls = [
        {
            "acl_action": "volume_view",
            "acl_type": view_acl_type,
            "userlist": view_acl_list if view_acl_type == "userlist" else [],
            "grouplist": view_acl_list if view_acl_type == "grouplist" else [],
        },
        {
            "acl_action": "volume_mount",
            "acl_type": mount_acl_type,
            "userlist": mount_acl_list if mount_acl_type == "userlist" else [],
            "grouplist": (
                mount_acl_list if mount_acl_type == "grouplist" else []
            ),
        },
        {
            "acl_action": "volume_edit",
            "acl_type": edit_acl_type,
            "userlist": edit_acl_list if edit_acl_type == "userlist" else [],
            "grouplist": edit_acl_list if edit_acl_type == "grouplist" else [],
        },
    ]
    payload = {
        "name": name,
        "description": description,
        "size": size,
        "acls": acls,
    }

    return request_adaboard(
        "volumes", method=requests.post, json=payload
    ).json()["id"]


def mount_volume_to_user_lab(
    volume_id: int,
    mount_path: str,
    user_id: Optional[str] = None,
    read_only: bool = False,
    from_root: bool = False,
) -> Response:
    """Create a new AdaLab Shared Volume mount to a specific AdaLab User's Lab.

    :param volume_id: The ID of the volume to mount.
    :type volume_id: int
    :param mount_path: The path to mount the volume to.
    :type mount_path: str
    :param user_id: The ID of the AdaLab User. If None, the current user is used.
    :type user_id: Optional[str]
    :param read_only: If True, mount the volume as read-only. Default is False.
    :type read_only: bool
    :param from_root: If True, mount the volume from the root. Default is False.
    :type from_root: bool

    :return: The response from the request.
    :rtype: Response
    """
    if user_id is None:
        user_id = request_adaboard(path="users/self").json()["user_id"]

    payload = {
        "mount_path": mount_path,
        "read_only": read_only,
        "from_root": from_root,
        "user_id": user_id,
    }
    return request_adaboard(
        f"volumes/{volume_id}/mount", method=requests.post, json=payload
    )


def update_volume(
    volume_id: int,
    name: Optional[str] = None,
    description: Optional[str] = None,
    size: Optional[int] = None,
    view_acl_type: Optional[str] = None,
    view_acl_list: list[str] = [],
    mount_acl_type: Optional[str] = None,
    mount_acl_list: list[str] = [],
    edit_acl_type: Optional[str] = None,
    edit_acl_list: list[str] = [],
) -> Response:
    """Update an AdaLab Shared Volume.

    :param volume_id: The ID of the volume to update.
    :type volume_id: int
    :param name: The new name of the volume.
    :type name: Optional[str]
    :param description: The new description of the volume.
    :type description: Optional[str]
    :param size: The new size of the volume in GB.
    :type size: Optional[int]
    :param view_acl_type: The new ACL type for viewing the volume.
    :type view_acl_type: Optional[str]
    :param view_acl_list: The new list of users or groups that can view the volume.
    :type view_acl_list: list[str]
    :param mount_acl_type: The new ACL type for mounting the volume.
    :type mount_acl_type: Optional[str]
    :param mount_acl_list: The new list of users or groups that can mount the volume.
    :type mount_acl_list: list[str]
    :param edit_acl_type: The new ACL type for editing the volume.
    :type edit_acl_type: Optional[str]
    :param edit_acl_list: The new list of users or groups that can edit the volume.
    :type edit_acl_list: list[str]

    :return: The response from the request.
    :rtype: Response
    """
    old_volume = get_volume(volume_id=volume_id)

    # Build ACLs payload
    old_acls = old_volume["acls"]
    acls = []

    for _acl_type, _acl_list, _acl_action in [
        (view_acl_type, view_acl_list, "volume_view"),
        (mount_acl_type, mount_acl_list, "volume_mount"),
        (edit_acl_type, edit_acl_list, "volume_edit"),
    ]:
        if _acl_type is not None:
            validate_acl("volume", False, _acl_type)
            acls.append(
                {
                    "acl_action": _acl_action,
                    "acl_type": _acl_type,
                    "userlist": _acl_list if _acl_type == "userlist" else [],
                    "grouplist": _acl_list if _acl_type == "grouplist" else [],
                }
            )
        else:
            old_view_acl = [
                x for x in old_acls if x["acl_action"] == _acl_action
            ][0]
            acls.append(
                {
                    "acl_action": old_view_acl["acl_action"],
                    "acl_type": old_view_acl["acl_type"],
                    "userlist": [
                        x["user_id"] for x in old_view_acl["userlist"]
                    ],
                    "grouplist": [
                        x["group_id"] for x in old_view_acl["grouplist"]
                    ],
                }
            )

    payload = {
        "name": name or old_volume["name"],
        "description": description or old_volume["description"],
        "size": size or old_volume["size"],
        "acls": acls,
    }
    return request_adaboard(
        f"volumes/{volume_id}", method=requests.put, json=payload
    )


def update_mounted_volume_to_user_lab(
    volume_id: int,
    user_id: Optional[str] = None,
    mount_path: Optional[str] = None,
    read_only: bool = False,
    from_root: bool = False,
) -> Response:
    """Update an AdaLab Shared Volume mount to a specific AdaLab User's Lab.

    :param volume_id: The ID of the volume to update.
    :type volume_id: int
    :param user_id: The ID of the AdaLab User. If None, the current user is used.
    :type user_id: Optional[str]
    :param mount_path: The new path to mount the volume to.
    :type mount_path: Optional[str]
    :param read_only: If True, mount the volume as read-only. Default is False.
    :type read_only: bool
    :param from_root: If True, mount the volume from the root. Default is False.
    :type from_root: bool

    :return: The response from the request.
    :rtype: Response
    """
    if user_id is None:
        user_id = request_adaboard(path="users/self").json()["user_id"]

    all_user_mounts = get_all_volume_mounts_for_user(user_id=user_id)
    spums = [x for x in all_user_mounts if x["volume_id"] == volume_id]
    assert len(spums) == 1, "User is not mounted to this volume"
    old_user_mount = spums[0]

    payload = {
        "mount_path": mount_path or old_user_mount["mount_path"],
        "read_only": read_only or old_user_mount["read_only"],
        "from_root": from_root or old_user_mount["from_root"],
        "user_id": user_id,
    }
    return request_adaboard(
        f"volumes/{volume_id}/mount", method=requests.put, json=payload
    )


def delete_volume(volume_id: int) -> Response:
    """Delete an AdaLab Shared Volume.

    :param volume_id: The ID of the volume to delete.
    :type volume_id: int
    :return: The response from the request.
    :rtype: Response
    """
    return request_adaboard(f"volumes/{volume_id}", method=requests.delete)


def unmount_volume_from_user_lab(
    volume_id: int, user_id: Optional[str] = None
) -> Response:
    """Unmount an AdaLab Shared Volume from a specific AdaLab User's Lab.

    :param volume_id: The ID of the volume to unmount.
    :type volume_id: int
    :param user_id: The ID of the AdaLab User. If None, the current user is used.
    :type user_id: Optional[str]

    :return: The response from the request.
    :rtype: Response
    """
    if user_id is None:
        user_id = request_adaboard(path="users/self").json()["user_id"]

    return request_adaboard(
        f"volumes/{volume_id}/mount",
        method=requests.delete,
        params={"user_id": user_id},
    )



import base64
import os

import requests

from .. import adaboard
from ..utils import validate_acl

ALLOWED_CARD_SOURCES = ["lab", "local"]
MAX_FILE_SIZE = 5000000
RUN_STATUS_CODES = {"1": "Pending", "2": "Running", "3": "OK", "4": "Error"}


def create_schedule(
    name: str,
    schedule: str,
    acl_type_view: str = "logged_in",
    acl_list_view: list[str] = [],
    acl_type_logs: str = "logged_in",
    acl_list_logs: list[str] = [],
    acl_type_edit: str = "logged_in",
    acl_list_edit: list[str] = [],
    acl_type_decrypt: str = "logged_in",
    acl_list_decrypt: list[str] = [],
    active: bool = True,
    aux_files: list[dict] = [{"source": "", "target": ""}],
    card_id: int = None,
    cleanup: bool = True,
    inputs: dict = {},
    concurrent: bool = True,
    kernel_id: int = None,
    notebook_file: str = "",
    owner_id: str = "",
    pool: str = "card-runner-low",
    post_run_script: str = "",
    pre_run_script: str = "",
    profile_id: int = None,
    options_ids: list[int] = [],
    runner_id: str = "",
    source: str = "lab",
    timeout: int = 3600,
    timezone: str = "Etc/UTC",
):
    """
    Create a new notebook schedule.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/create_schedule.ipynb) to test this function or build upon it.

    :param name: The schedule's name.
    :type name: str
    :param schedule: The schedule time string in cron format.
    :type schedule: str
    :param acl_type_view: The ACL type for viewing the schedule. Defaults to "logged_in".
    :type acl_type_view: str, optional
    :param acl_list_view: The list of users or groups allowed to view the schedule. Defaults to an empty list.
    :type acl_list_view: list[str], optional
    :param acl_type_logs: The ACL type for viewing the schedule including logs. Defaults to "logged_in".
    :type acl_type_logs: str, optional
    :param acl_list_logs: The list of users or groups allowed to view the schedule including logs. Defaults to an empty list.
    :type acl_list_logs: list[str], optional
    :param acl_type_edit: The ACL type for editing the schedule. Defaults to "logged_in".
    :type acl_type_edit: str, optional
    :param acl_list_edit: The list of users or groups allowed to edit the schedule. Defaults to an empty list.
    :type acl_list_edit: list[str], optional
    :param acl_type_decrypt: The ACL type for editing the schedule and decrypting its secrets. Defaults to "logged_in".
    :type acl_type_decrypt: str, optional
    :param acl_list_decrypt: The list of users or groups allowed to edit the schedule and decrypt its secrets. Defaults to an empty list.
    :type acl_list_decrypt: list[str], optional
    :param active: Flag whether the schedule is active. Defaults to True.
    :type active: bool, optional
    :param aux_files: A list of auxiliary files to include in the schedule.
    :type aux_files: list[dict], optional
    :param card_id: The ID of the card from which to create the schedule. Defaults to None.
    :type card_id: int, optional
    :param cleanup: Flag whether to clean up resources after the schedule. Defaults to True.
    :type cleanup: bool, optional
    :param inputs: A dictionary of input parameters for the schedule. Defaults to an empty dictionary.
    :type inputs: dict, optional
    :param concurrent: Flag whether the schedule can run concurrently. Defaults to True.
    :type concurrent: bool, optional
    :param kernel_id: The ID of the kernel to be used when running the schedule. Defaults to None.
    :type kernel_id: int, optional
    :param notebook_file: The path to the notebook file. Defaults to an empty string.
    :type notebook_file: str, optional
    :param owner_id: The user ID of the person to own the schedule. Defaults to an empty string.
    :type owner_id: str, optional
    :param pool: The execution pool. Defaults to "card-runner-low".
    :type pool: str, optional
    :param post_run_script: A post-run script to execute. Defaults to an empty string.
    :type post_run_script: str, optional
    :param pre_run_script: A pre-run script to execute. Defaults to an empty string.
    :type pre_run_script: str, optional
    :param profile_id: The ID of the Lab profile to use when running the notebook. Defaults to None.
    :type profile_id: int, optional
    :param options_ids: A list of Lab option IDs to use when running the notebook. Defaults to an empty list.
    :type options_ids: list[int], optional
    :param runner_id: The user ID of the person running the notebook schedule. Defaults to an empty string.
    :type runner_id: str, optional
    :param source: The source of the notebook file. Defaults to "lab".
    :type source: str, optional
    :param timeout: The maximum time (in seconds) before killing the run. Defaults to 3600.
    :type timeout: int, optional
    :param timezone: The timezone code for the schedule. Defaults to "Etc/UTC".
    :type timezone: str, optional
    :return: The schedule's ID.
    :rtype: int
    """

    assert (
        card_id or notebook_file
    ), "Either card_id or notebook_file must be provided."
    if notebook_file:
        assert notebook_file.lower().endswith(
            ".ipynb"
        ), "Notebook file must be a .ipynb file."
        assert (
            source.lower() in ALLOWED_CARD_SOURCES
        ), "Files source must be 'lab' or 'local'."
        assert all(
            issubclass(type(x), dict) for x in aux_files
        ), "Auxiliary files must be a list of dictionaries."
        assert all(
            ("source" in d and "target" in d) for d in aux_files
        ), "Auxiliary files must have keys 'source' and 'target'."

    validate_acl(
        "schedule",
        False,
        acl_type_view,
        acl_type_logs,
        acl_type_edit,
        acl_type_decrypt,
    )

    # Collect ACL configuration
    acl_config = [
        {
            "acl_action": "schedule_view",
            "acl_type": acl_type_view,
            "userlist": acl_list_view,
            "grouplist": acl_list_view,
        },
        {
            "acl_action": "schedule_view_logs",
            "acl_type": acl_type_logs,
            "userlist": acl_list_logs,
            "grouplist": acl_list_logs,
        },
        {
            "acl_action": "schedule_edit",
            "acl_type": acl_type_edit,
            "userlist": acl_list_edit,
            "grouplist": acl_list_edit,
        },
        {
            "acl_action": "schedule_edit_privileged",
            "acl_type": acl_type_decrypt,
            "userlist": acl_list_decrypt,
            "grouplist": acl_list_decrypt,
        },
    ]
    # Collect run configuration
    run_config = {
        "inputs": inputs,
        "options_ids": options_ids,
        "profile_id": profile_id,
        "scripts": {
            "pre_run_script": pre_run_script,
            "post_run_script": post_run_script,
        },
    }

    # Add content depending on user input
    content = {}
    if card_id:
        tmp = adaboard.request_adaboard(
            f"cards/{card_id}", params={"incl_content": True}
        ).json()
        content["content"] = tmp["content"]
    else:
        if source.lower() == "lab":
            # Check file size before actually pulling it from the Lab
            try:
                response_props = adaboard.request_adaboard(
                    path="jupyter/files/notebook/content/",
                    params={"path": os.path.dirname(notebook_file)},
                ).json()
                props = [
                    x
                    for x in response_props
                    if x["name"] == os.path.basename(notebook_file)
                ][0]
            except IndexError as e:
                raise IndexError(
                    f"Notebook file '{notebook_file}' not found in the Lab."
                ) from e
            assert (
                props["size"] <= MAX_FILE_SIZE
            ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
            nb_file_bin = adaboard.request_adaboard(
                path="jupyter/files/notebook/content/fetch",
                params={"path": notebook_file},
            ).json()["content"]
        else:
            # Check file size before actually opening it
            assert (
                os.path.getsize(notebook_file) <= MAX_FILE_SIZE
            ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
            with open(notebook_file, "rb") as f:
                nb_file_bin = base64.b64encode(f.read()).decode("utf-8")
        content["content"] = [
            {
                "content": nb_file_bin,
                "content_name": os.path.basename(notebook_file),
                "contenttype_id": 1,
            }
        ]
    # Auxiliary files
    # Collect all files within the specified paths
    if aux_files[0]["source"]:
        # In the Lab, paths are relative to the user's home directory
        if source.lower() == "lab":
            # Collect all files within the specified paths
            all_aux_files = []
            for aux_file in aux_files:
                # Directories end with "/"; otherwise they are considered files; no nesting allowed
                if aux_file["source"].endswith("/"):
                    dir_cont = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": aux_file["source"]},
                    ).json()
                    dir_paths = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in dir_cont
                        if x["type"] == "file"
                    ]
                    all_aux_files.extend(dir_paths)
                else:
                    response_props = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": os.path.dirname(aux_file["source"])},
                    ).json()
                    file_path = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in response_props
                        if x["name"] == os.path.basename(aux_file["source"])
                    ]
                    all_aux_files.extend(file_path)
            # Now we have all the files with their paths and sizes in a single list
            for aux_file in all_aux_files:
                # Check file size before actually pulling it from the Lab
                assert (
                    aux_file["size"] <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                # "content" is the binary; "content_name" determines the target folder
                aux_file_bin = adaboard.request_adaboard(
                    path="jupyter/files/notebook/content/fetch",
                    params={"path": aux_file["source"]},
                ).json()["content"]
                aux_file_name = f"{aux_file['target']}{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                    # "path": aux_file["source"],
                }
                content["content"].append(tmp)

        # In the local machine, paths are absolute; nesting is allowed
        elif source.lower() == "local":
            all_aux_files = []
            for aux_file in aux_files:
                if os.path.isfile(aux_file["source"]):
                    all_aux_files.append(aux_file)
                elif os.path.isdir(aux_file["source"]):
                    for root, dirs, files in os.walk(aux_file["source"]):
                        for file in files:
                            all_aux_files.append(
                                {
                                    "source": f"{root}/{file}",
                                    "target": f"{root}/".replace(
                                        aux_file["source"], aux_file["target"]
                                    ),
                                }
                            )
            # Now we have all the files with their paths in a single list
            for aux_file in all_aux_files:
                # Check file sizes before actually opening them
                assert (
                    os.path.getsize(aux_file["source"]) <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                with open(aux_file["source"], "rb") as f:
                    aux_file_bin = base64.b64encode(f.read()).decode("utf-8")
                # "content" is the binary; "content_name" determines the target folder
                aux_file_name = f"{aux_file['target']}/{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                }
                content["content"].append(tmp)

    # Build request payload with user options
    payload = {
        "owner_id": owner_id or adaboard.get_user()["user_id"],
        "runner_id": runner_id or adaboard.get_user()["user_id"],
        "name": name,
        "timezone": timezone,
        "cron": schedule,
        "timeout": timeout,
        "pool": pool,
        "active": active,
        "concurrent": concurrent,
        "cleanup": cleanup,
        "acls": acl_config,
        "run_config": run_config,
        "kernel_metadata_id": kernel_id,
    }
    # Attach schedule content to payload
    payload.update(content)

    response = adaboard.request_adaboard(
        "schedules/v2", method=requests.post, json=payload
    ).json()

    return response["id"]


def delete_run(schedule_id: int, run_id: int) -> None:
    """
    Delete a specific schedule run.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/delete_run.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID.
    :type schedule_id: int
    :param run_id: The run's ID.
    :type run_id: int
    :return: Nothing.
    :rtype: None
    """

    adaboard.request_adaboard(
        path=f"schedules/v2/runs/{schedule_id}/{run_id}",
        method=requests.delete,
    )


def delete_schedule(schedule_id: int) -> None:
    """
    Delete a specific notebook schedule.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/delete_schedule.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID.
    :type schedule_id: int
    :return: Nothing.
    :rtype: None
    """

    adaboard.request_adaboard(
        path=f"schedules/v2/{schedule_id}", method=requests.delete
    )


def edit_schedule(  # noqa: C901
    schedule_id: int,
    name: str = None,
    schedule: str = None,
    acl_type_view: str = None,
    acl_list_view: list[str] = None,
    acl_type_logs: str = None,
    acl_list_logs: list[str] = None,
    acl_type_edit: str = None,
    acl_list_edit: list[str] = None,
    acl_type_decrypt: str = None,
    acl_list_decrypt: list[str] = None,
    active: bool = None,
    aux_files: list[dict] = None,
    cleanup: bool = None,
    inputs: dict = None,
    concurrent: bool = None,
    keep_aux_files: bool = True,
    kernel_id: int = None,
    notebook_file: str = None,
    owner_id: str = None,
    pool: str = None,
    post_run_script: str = None,
    pre_run_script: str = None,
    profile_id: int = None,
    options_ids: list[int] = None,
    runner_id: str = None,
    source: str = "lab",
    timeout: int = None,
    timezone: str = None,
):
    """
    Edit an existing notebook schedule. Note that unspecified fields will not be changed.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/edit_schedule.ipynb) to test this function or build upon it.

    :param name: The name of the schedule. Defaults to None
    :type name: str, optional
    :param schedule: The schedule time string in cron format. Defaults to None
    :type schedule: str, optional
    :param acl_type_view: The ACL type for viewing the schedule. Defaults to None
    :type acl_type_view: str, optional
    :param acl_list_view: The list of users or groups allowed to view the schedule. Defaults to None
    :type acl_list_view: list[str], optional
    :param acl_type_logs: The ACL type for viewing the schedule including logs. Defaults to None
    :type acl_type_logs: str, optional
    :param acl_list_logs: The list of users or groups allowed to view the schedule including logs. Defaults to None
    :type acl_list_logs: list[str], optional
    :param acl_type_edit: The ACL type for editing the schedule. Defaults to None
    :type acl_type_edit: str, optional
    :param acl_list_edit: The list of users or groups allowed to edit the schedule. Defaults to None
    :type acl_list_edit: list[str], optional
    :param acl_type_decrypt: The ACL type for editing the schedule and decrypting its secrets. Defaults to None
    :type acl_type_decrypt: str, optional
    :param acl_list_decrypt: The list of users or groups allowed to edit the schedule and decrypt its secrets. Defaults to None
    :type acl_list_decrypt: list[str], optional
    :param active: Flag whether the schedule is active. Defaults to None
    :type active: bool, optional
    :param aux_files: A list of auxiliary files to include in the schedule. Defaults to None
    :type aux_files: list[dict], optional
    :param card_id: The ID of the card from which to create the schedule. Defaults to None
    :type card_id: int, optional
    :param cleanup: Flag whether to clean up resources after the schedule. Defaults to None
    :type cleanup: bool, optional
    :param inputs: A dictionary of input parameters for the schedule. Defaults to None
    :type inputs: dict, optional
    :param concurrent: Flag whether the schedule can run concurrently. Defaults to None
    :type concurrent: bool, optional
    :param keep_aux_files: Flag whether to keep the existing auxiliary files. Defaults to True
    :type keep_aux_files: bool, optional
    :param kernel_id: The ID of the kernel to be used when running the schedule. Defaults to None
    :type kernel_id: int, optional
    :param notebook_file: The path to the notebook file. Defaults to None
    :type notebook_file: str, optional
    :param owner_id: The user ID of the person to own the schedule. Defaults to None
    :type owner_id: str, optional
    :param pool: The execution pool. Defaults to None
    :type pool: str, optional
    :param post_run_script: A post-run script to execute. Defaults to None
    :type post_run_script: str, optional
    :param pre_run_script: A pre-run script to execute. Defaults to None
    :type pre_run_script: str, optional
    :param profile_id: The ID of the Lab profile to use when running the notebook. Defaults to None
    :type profile_id: int, optional
    :param options_ids: A list of Lab option IDs to use when running the notebook. Defaults to None
    :type options_ids: list[int], optional
    :param runner_id: The user ID of the person running the notebook schedule. Defaults to None
    :type runner_id: str, optional
    :param source: The source of the schedule files. Defaults to "lab"
    :type source: str, optional
    :param timeout: The maximum time (in seconds) before killing the run. Defaults to None
    :type timeout: int, optional
    :param timezone: The timezone code for the schedule. Defaults to None
    :type timezone: str, optional
    :return: Nothing
    :rtype: None
    """
    if notebook_file:
        assert notebook_file.lower().endswith(
            ".ipynb"
        ), "Notebook file must be a .ipynb file."
        assert (
            source.lower() in ALLOWED_CARD_SOURCES
        ), "Files source must be 'lab' or 'local'."
    if aux_files:
        assert all(
            issubclass(type(x), dict) for x in aux_files
        ), "Auxiliary files must be a list of dictionaries."
        assert all(
            ("source" in d and "target" in d) for d in aux_files
        ), "Auxiliary files must have keys 'source' and 'target'."

    validate_acl(
        "schedule",
        True,
        acl_type_view,
        acl_type_logs,
        acl_type_edit,
        acl_type_decrypt,
    )

    # Collect new schedule configuration
    schedule_config = {
        "owner_id": owner_id,
        "runner_id": runner_id,
        "name": name,
        "timezone": timezone,
        "cron": schedule,
        "timeout": timeout,
        "pool": pool,
        "active": active,
        "concurrent": concurrent,
        "cleanup": cleanup,
        "kernel_metadata_id": kernel_id,
    }

    # Fetch old configuration
    tmp = adaboard.request_adaboard(
        f"schedules/v2/{schedule_id}",
        method=requests.get,
        params={"with_content": True},
    ).json()
    tmp.pop("kernel")
    tmp.pop("owner")
    tmp.pop("runner")
    tmp.pop("last_scheduled_ts")
    tmp.pop("next_scheduled_ts")
    # Reformat and load old ACLs
    old_acls = [
        {
            "acl_action": x["acl_action"],
            "acl_type": x["acl"]["acl_type"],
            "userlist": [
                y["user_id"]
                for y in x["acl"]["userlist"]
                if "user_id" in y.keys()
            ],
            "grouplist": [
                y["user_id"]
                for y in x["acl"]["grouplist"]
                if "user_id" in y.keys()
            ],
        }
        for x in tmp["acls"]
    ]

    old_config = tmp
    old_config.update({"acls": old_acls})
    # Update ACLs
    acl_config = [
        {
            "acl_action": "schedule_view",
            "acl_type": (
                acl_type_view
                if acl_type_view
                else old_config["acls"][0]["acl_type"]
            ),
            "userlist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["userlist"]
            ),
            "grouplist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["grouplist"]
            ),
        },
        {
            "acl_action": "schedule_view_logs",
            "acl_type": (
                acl_type_logs
                if acl_type_logs
                else old_config["acls"][1]["acl_type"]
            ),
            "userlist": (
                acl_list_logs
                if acl_list_logs
                else old_config["acls"][1]["userlist"]
            ),
            "grouplist": (
                acl_list_logs
                if acl_list_logs
                else old_config["acls"][1]["grouplist"]
            ),
        },
        {
            "acl_action": "schedule_edit",
            "acl_type": (
                acl_type_edit
                if acl_type_edit
                else old_config["acls"][2]["acl_type"]
            ),
            "userlist": (
                acl_list_edit
                if acl_list_edit
                else old_config["acls"][2]["userlist"]
            ),
            "grouplist": (
                acl_list_edit
                if acl_list_edit
                else old_config["acls"][2]["grouplist"]
            ),
        },
        {
            "acl_action": "schedule_edit_privileged",
            "acl_type": (
                acl_type_decrypt
                if acl_type_decrypt
                else old_config["acls"][3]["acl_type"]
            ),
            "userlist": (
                acl_list_decrypt
                if acl_list_decrypt
                else old_config["acls"][3]["userlist"]
            ),
            "grouplist": (
                acl_list_decrypt
                if acl_list_decrypt
                else old_config["acls"][3]["grouplist"]
            ),
        },
    ]
    schedule_config.update({"acls": acl_config})
    # Update run configuration
    run_config = {
        "inputs": inputs if inputs else old_config["run_config"]["inputs"],
        "options_ids": (
            options_ids
            if options_ids is not None
            else old_config["run_config"]["options_ids"]
        ),
        "profile_id": (
            profile_id
            if profile_id is not None
            else old_config["run_config"]["profile_id"]
        ),
        "scripts": {
            "pre_run_script": (
                pre_run_script
                if pre_run_script is not None
                else old_config["run_config"]["scripts"]["pre_run_script"]
            ),
            "post_run_script": (
                post_run_script
                if post_run_script is not None
                else old_config["run_config"]["scripts"]["post_run_script"]
            ),
        },
    }
    schedule_config.update({"run_config": run_config})

    # Update content
    # First, the main Notebook
    if notebook_file:
        if source.lower() == "lab":
            # Check file size before actually pulling it from the Lab
            try:
                response_props = adaboard.request_adaboard(
                    path="jupyter/files/notebook/content/",
                    params={"path": os.path.dirname(notebook_file)},
                ).json()
                props = [
                    x
                    for x in response_props
                    if x["name"] == os.path.basename(notebook_file)
                ][0]
            except IndexError as e:
                raise IndexError(
                    f"Notebook file '{notebook_file}' not found in the Lab."
                ) from e
            assert (
                props["size"] <= MAX_FILE_SIZE
            ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
            nb_file_bin = adaboard.request_adaboard(
                path="jupyter/files/notebook/content/fetch",
                params={"path": notebook_file},
            ).json()["content"]
        else:
            # Check file size before actually opening it
            assert (
                os.path.getsize(notebook_file) <= MAX_FILE_SIZE
            ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
            with open(notebook_file, "rb") as f:
                nb_file_bin = base64.b64encode(f.read()).decode("utf-8")

        content = [
            {
                "content": nb_file_bin,
                "content_name": os.path.basename(notebook_file),
                "contenttype_id": 1,
            }
        ]
    else:
        content = [
            x for x in old_config["content"] if x["contenttype_id"] == 1
        ]

    # Next, the auxiliary files
    # Start by collecting or ditching the old ones
    if keep_aux_files:
        old_aux_files = [
            x for x in old_config["content"] if x["contenttype_id"] == 2
        ]
        content.extend(old_aux_files)
    # Then, add the new ones
    if aux_files:
        # In the Lab, paths are relative to the user's home directory
        if source.lower() == "lab":
            # Collect all files within the specified paths
            all_aux_files = []
            for aux_file in aux_files:
                # Directories end with "/"; otherwise they are considered files; no nesting allowed
                if aux_file["source"].endswith("/"):
                    dir_cont = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": aux_file["source"]},
                    ).json()
                    dir_paths = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in dir_cont
                        if x["type"] == "file"
                    ]
                    all_aux_files.extend(dir_paths)
                else:
                    response_props = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": os.path.dirname(aux_file["source"])},
                    ).json()
                    file_path = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in response_props
                        if x["name"] == os.path.basename(aux_file["source"])
                    ]
                    all_aux_files.extend(file_path)
            # Now we have all the files with their paths and sizes in a single list
            for aux_file in all_aux_files:
                # Check file size before actually pulling it from the Lab
                assert (
                    aux_file["size"] <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                # "content" is the binary; "content_name" determines the target folder
                aux_file_bin = adaboard.request_adaboard(
                    path="jupyter/files/notebook/content/fetch",
                    params={"path": aux_file["source"]},
                ).json()["content"]
                aux_file_name = f"{aux_file['target']}{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                }
                content.append(tmp)

        # In the local machine, paths are absolute; nesting is allowed
        elif source.lower() == "local":
            all_aux_files = []
            for aux_file in aux_files:
                if os.path.isfile(aux_file["source"]):
                    all_aux_files.append(aux_file)
                elif os.path.isdir(aux_file["source"]):
                    for root, dirs, files in os.walk(aux_file["source"]):
                        for file in files:
                            all_aux_files.append(
                                {
                                    "source": f"{root}/{file}",
                                    "target": f"{root}/".replace(
                                        aux_file["source"], aux_file["target"]
                                    ),
                                }
                            )
            # Now we have all the files with their paths in a single list
            for aux_file in all_aux_files:
                # Check file sizes before actually opening them
                assert (
                    os.path.getsize(aux_file["source"]) <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                with open(aux_file["source"], "rb") as f:
                    aux_file_bin = base64.b64encode(f.read()).decode("utf-8")
                # "content" is the binary; "content_name" determines the target folder
                aux_file_name = f"{aux_file['target']}/{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                }
                content.append(tmp)

    schedule_config.update({"content": content})

    # Finally, build the request payload based on new and old configuration
    payload = {
        k: v if v is not None else old_config[k]
        for k, v in schedule_config.items()
    }

    # return payload
    adaboard.request_adaboard(
        f"schedules/v2/{schedule_id}", method=requests.put, json=payload
    )


def get_all_schedules(include_inactive: bool = False) -> list[dict]:
    """
    Retrieve a list of schedules for all users.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_all_schedules.ipynb) to test this function or build upon it.

    :param include_inactive: Flag whether inactive schedules should be included. Defaults to False.
    :type include_inactive: bool, optional
    :return: A list of notebook schedules in dictionary form.
    :rtype: list[dict]
    """

    all_schedules = adaboard.get_all_pages(
        "schedules/v2", params={"only_active": not include_inactive}
    )
    clean_schedules = [__extract_schedule_data(x) for x in all_schedules]
    return clean_schedules


def get_card_schedules(
    card_id: int, include_inactive: bool = False
) -> list[dict]:
    """
    Retrieve a list of schedules for a specific card.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_card_schedules.ipynb) to test this function or build upon it.

    :param card_id: The card's ID.
    :type card_id: int
    :param include_inactive: Flag whether inactive schedules should be included. Defaults to False.
    :type include_inactive: bool, optional
    :return: A list of notebook schedules in dictionary form.
    :rtype: list[dict]
    """

    all_schedules = adaboard.get_all_pages(
        "schedules/v2",
        params={"only_active": not include_inactive},
    )

    cards_schedules = [
        __extract_schedule_data(x)
        for x in all_schedules
        if x["card_id"] == card_id
    ]

    return cards_schedules


def get_pool_stats() -> list[dict[str, int | str | list]]:
    """
    Retrieve statistics about all run pools.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_pool_stats.ipynb) to test this function or build upon it.
    """

    return adaboard.request_adaboard(
        path="schedules/v2/runs/stats", method=requests.get
    ).json()


def get_run_info(schedule_id: int, run_id: int) -> dict:
    """
    Retrieve information about a specific run.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_run_info.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID.
    :type schedule_id: int
    :param run_id: The run's ID.
    :type run_id: int
    :return: Information about the run.
    :rtype: dict
    """

    run_info = adaboard.request_adaboard(
        f"schedules/v2/runs/{schedule_id}/{run_id}",
    ).json()
    run_info.pop("user")
    run_info["status"] = RUN_STATUS_CODES[str(run_info["status"])]
    return run_info


def get_run_logs(schedule_id: int, run_id: int) -> dict:
    """
    Retrieve the execution and system logs of a specific run.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_run_logs.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID.
    :type schedule_id: int
    :param run_id: The run's ID.
    :type run_id: int
    :return: Information about the run.
    :rtype: dict
    """

    return adaboard.request_adaboard(
        f"schedules/v2/runs/{schedule_id}/{run_id}/logs",
    ).json()


def get_runs_overview(
    schedule_id: int = None, owner_id: int = None
) -> list[dict]:
    """
    Retrieve a list of all schedule runs.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_runs_overview.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID. Defaults to all.
    :type schedule_id: int, optional
    :param owner_id: The schedule's owner ID. Defaults to all.
    :type owner_id: int, optional
    """

    all_runs = adaboard.get_all_pages(
        "schedules/v2/runs/overview",
        params={"user_id": owner_id, "schedule_id": schedule_id},
    )
    try:
        all_runs.pop("owner_image_id")
        for runs in all_runs:
            for run in runs["runs"]:
                run["status"] = RUN_STATUS_CODES[str(run["status"])]
    except (TypeError, KeyError):
        pass
    return all_runs


def get_schedule(
    schedule_id: int,
) -> dict[str, str | int | bool | list | dict]:
    """
    Retrieve a specific notebook schedule.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_schedule.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID.
    :type schedule_id: int
    :return: A list of notebook schedules in dictionary form.
    :rtype: list[dict]
    """

    schedule = adaboard.request_adaboard(
        path=f"schedules/v2/{schedule_id}", method=requests.get
    ).json()
    clean_schedule = __extract_schedule_data(schedule)
    return clean_schedule


def get_schedule_id(name: str, owner_id: str) -> list[int]:
    """
    Find the ID of a specific schedule.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_schedule_id.ipynb) to test this function or build upon it.

    :param name: The schedule's name.
    :type name: str
    :param owner_id: The schedule's owner ID.
    :type owner_id: str
    :return: A list of schedule IDs that match the search parameters.
    :rtype: list[int]
    """

    all_schedules = adaboard.get_all_pages(
        "schedules/v2",
        params={"owner_id": owner_id},
    )

    schedule_id = [
        x["schedule_id"]
        for x in all_schedules
        if x["owner_id"] == owner_id and x["name"] == name
    ]
    if not schedule_id:
        raise ValueError(
            f"No schedule found with name '{name}' from user '{owner_id}'."
        )
    return schedule_id


def get_user_schedules(
    user_id: str = "", include_inactive: bool = False
) -> list[dict]:
    """
    Retrieve a list of schedules for a specific user.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/get_user_schedules.ipynb) to test this function or build upon it.

    :param user_id: The user's ID. Defaults to the current user.
    :type user_id: str, optional
    :param include_inactive: Flag whether inactive schedules should be included. Defaults to False.
    :type include_inactive: bool, optional
    :return: A list of notebook schedules in dictionary form.
    :rtype: list[dict]
    """
    user_id = user_id or adaboard.get_user()["user_id"]

    get_user_schedules = adaboard.get_all_pages(
        "schedules/v2",
        params={"user_id": user_id, "only_active": not include_inactive},
    )
    clean_schedules = [__extract_schedule_data(x) for x in get_user_schedules]
    return clean_schedules


def start_run(schedule_id: int) -> int:
    """
    Trigger a run of a specific notebook schedule.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/start_run.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID.
    :type schedule_id: int
    :return: Schedule run's ID
    :rtype: int
    """

    res = adaboard.request_adaboard(
        f"schedules/v2/runs/{schedule_id}/trigger", method=requests.post
    ).json()

    return res["id"]


def stop_run(schedule_id: int, run_id: int) -> None:
    """
    Stop a specific run of a notebook schedule.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/schedules/stop_run.ipynb) to test this function or build upon it.

    :param schedule_id: The schedule's ID.
    :type schedule_id: int
    :param run_id: The run's ID.
    :type run_id: int
    :return: Nothing.
    :rtype: None
    """

    adaboard.request_adaboard(
        f"schedules/v2/runs/{schedule_id}/{run_id}/stop", method=requests.post
    ).json()


def __extract_schedule_data(raw_data: dict) -> dict:
    """
    Helper function to discard the unnecessary fields from the raw schedule data.

    :param raw_data: Raw schedule data.
    :type raw_data: dict
    :return: Clean schedule data
    :rtype: dict
    """

    raw_data.pop("owner")
    raw_data.pop("runner")
    raw_data.pop("kernel")
    acls = [
        {
            "acl_action": x["acl_action"],
            "acl_type": x["acl"]["acl_type"],
            "userlist": [
                y["user_id"]
                for y in x["acl"]["userlist"]
                if "user_id" in y.keys()
            ],
            "grouplist": [
                y["user_id"]
                for y in x["acl"]["grouplist"]
                if "user_id" in y.keys()
            ],
        }
        for x in raw_data["acls"]
    ]
    raw_data["acls"] = acls
    return raw_data



"The Cards sub-package exposes the core integrations of the cards in the AdaLab Gallery."

from .cards import (
    create_card,
    create_card_group,
    create_url_card,
    create_url_subtype,
    delete_card,
    delete_card_group,
    delete_url_subtype,
    edit_card,
    edit_card_group,
    edit_url_card,
    edit_url_subtype,
    expose_card,
    get_card_contents,
    get_card_issues,
    get_card_types,
    get_card_types_stats,
    get_cards,
    get_url_subtypes,
    hide_card,
    launch_card,
    set_card_visibility,
    toggle_card_favorite,
    toggle_card_review,
)

__all__ = [
    "create_card",
    "create_card_group",
    "create_url_card",
    "create_url_subtype",
    "delete_card",
    "delete_card_group",
    "delete_url_subtype",
    "edit_card",
    "edit_card_group",
    "edit_url_card",
    "edit_url_subtype",
    "expose_card",
    "get_card_contents",
    "get_card_issues",
    "get_card_types",
    "get_card_types_stats",
    "get_url_subtypes",
    "get_cards",
    "hide_card",
    "launch_card",
    "set_card_visibility",
    "toggle_card_favorite",
    "toggle_card_review",
]
__title__ = "adalib Cards"



"""The utils sub-package includes some non-API related functionalities in AdaLab."""

from .acl import validate_acl
from .volumes import validate_volume_mount

__all__ = [
    "validate_acl",
    "validate_volume_mount",
]
__title__ = "adalib utils"



"""The Schedules sub-package exposes the core integrations of the notebook schedules in the AdaLab Gallery.
"""

from .schedules import (
    create_schedule,
    delete_run,
    delete_schedule,
    edit_schedule,
    get_all_schedules,
    get_card_schedules,
    get_pool_stats,
    get_run_info,
    get_run_logs,
    get_runs_overview,
    get_schedule,
    get_schedule_id,
    get_user_schedules,
    start_run,
    stop_run,
)

__all__ = [
    "create_schedule",
    "delete_run",
    "delete_schedule",
    "edit_schedule",
    "get_all_schedules",
    "get_card_schedules",
    "get_pool_stats",
    "get_run_info",
    "get_run_logs",
    "get_runs_overview",
    "get_schedule",
    "get_schedule_id",
    "get_user_schedules",
    "start_run",
    "stop_run",
]

__title__ = "adalib Schedules"



import base64
import json
import os
import re
from typing import Union

import requests

from .. import adaboard
from ..keywords import create_keywords
from ..utils import validate_acl

ALLOWED_CARD_SOURCES = ["lab", "local"]
ALLOWED_CARD_TYPES = {"notebook": 1, "voila": 2}
MAX_FILE_SIZE = 5000000


def _create_keywords_if_not_exist(keywords: list[str]) -> dict[str, int]:
    """Helper function to create keywords if they do not exist in the Gallery.

    :param keywords: the list of keywords to be created
    :type keywords: list
    :return: the dictionary of all keywords and their IDs
    :rtype: dict[str, int]
    """
    all_kws = {
        x["keyword"]: x["keyword_id"]
        for x in adaboard.request_adaboard(
            path="keywords", method=requests.get
        ).json()
    }
    for kw in keywords:
        if kw not in all_kws:
            new_kw = create_keywords(new_keywords=kw)[0]
            all_kws[kw] = new_kw["keyword_id"]

    return all_kws


def create_card(  # noqa: C901
    name: str,
    description: str,
    cardtype: str,
    notebook_file: str,
    acl_type_view: str = "public",
    acl_list_view: list[str] = [],
    acl_type_launch: str = "public",
    acl_list_launch: list[str] = [],
    acl_type_edit: str = "public",
    acl_list_edit: list[str] = [],
    aux_files: list[dict[str, str]] = [{"source": "", "target": ""}],
    coauthors: list[str] = [],
    include_kernel: bool = True,
    keywords: list[str] = [],
    picture_id: int = 0,
    reviewers: list[str] = [],
    source: str = "lab",
) -> int:
    """
    Create a card in the AdaLab Gallery. Note that only Notebook and Voilà cards are supported.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/create_card.ipynb) to test this function or build upon it.

    :param name: the name of the card
    :type name: str
    :param description: the card description (tip: supports Markdown)
    :type description: str
    :param cardtype: the type of card (tip: must be "notebook" or "voila")
    :type cardtype: str
    :param notebook_file: the path to the main notebook file to be added to the card
    :type notebook_file: str
    :param acl_type_view: the ACL type for viewing the card, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_view: str, optional
    :param acl_list_view: the list of users/groups allowed to view the card, defaults to []
    :type acl_list_view: list, optional
    :param acl_type_launch: the ACL type for launching the card, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_launch: str, optional
    :param acl_list_launch: the list of users/groups allowed to launch the card, defaults to []
    :type acl_list_launch: list, optional
    :param acl_type_edit: the ACL type for editing the card, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_edit: str, optional
    :param acl_list_edit: the list of users/groups allowed to edit the card, defaults to []
    :type acl_list_edit: list, optional
    :param aux_files: the list of paths to the auxiliary files to be added to the card, defaults to {}
    :type aux_files: dict, optional
    :param coauthors: the other card authors, defaults to []
    :type coauthors: list, optional
    :param include_kernel: whether to link the Notebook with the kernel used during its development, defaults to True
    :type include_kernel: bool, optional
    :param keywords: the keywords related to the card, defaults to [] (tip: must exist in the Gallery)
    :type keywords: list, optional
    :param picture_id: the ID of the picture to be used for the card, defaults to 0
    :type picture_id: int, optional
    :param reviewers: the list of users who will review the card, defaults to []
    :type reviewers: list, optional
    :param source: the source of the card files, defaults to "lab" (tip: must be "lab" or "local")
    :type source: str, optional
    :return: the ID of the card
    :rtype: int
    """

    assert (
        cardtype.lower() in ALLOWED_CARD_TYPES
    ), "Card type must be 'notebook' or 'voila'."
    assert notebook_file.lower().endswith(
        ".ipynb"
    ), "Notebook file must be a .ipynb file."
    assert (
        source.lower() in ALLOWED_CARD_SOURCES
    ), "Files source must be 'lab' or 'local'."
    validate_acl("card", False, acl_type_view, acl_type_launch, acl_type_edit)
    assert all(
        issubclass(type(x), dict) for x in aux_files
    ), "Auxiliary files must be a list of dictionaries."
    assert all(
        ("source" in d and "target" in d) for d in aux_files
    ), "Auxiliary files must have keys 'source' and 'target'."

    cardtype_id = ALLOWED_CARD_TYPES[cardtype.lower()]

    keywords_id = []
    if keywords:
        kws = _create_keywords_if_not_exist(keywords=keywords)
        for kw in keywords:
            keywords_id.append(kws[kw])

    # Main notebook file; read binary from Lab or local machine
    if source.lower() == "lab":
        # Check file size before actually pulling it from the Lab
        try:
            response_props = adaboard.request_adaboard(
                path="jupyter/files/notebook/content/",
                params={"path": os.path.dirname(notebook_file)},
            ).json()
            props = [
                x
                for x in response_props
                if x["name"] == os.path.basename(notebook_file)
            ][0]
        except IndexError as e:
            raise IndexError(
                f"Notebook file '{notebook_file}' not found in the Lab."
            ) from e
        assert (
            props["size"] <= MAX_FILE_SIZE
        ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
        nb_file_bin = adaboard.request_adaboard(
            path="jupyter/files/notebook/content/fetch",
            params={"path": notebook_file},
        ).json()["content"]
    else:
        # Check file size before actually opening it
        assert (
            os.path.getsize(notebook_file) <= MAX_FILE_SIZE
        ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
        with open(notebook_file, "rb") as f:
            nb_file_bin = base64.b64encode(f.read()).decode("utf-8")

    if not include_kernel:
        nb_file_dict = json.loads(
            base64.b64decode(nb_file_bin).decode("utf-8")
        )
        nb_file_dict["metadata"]["kernelspec"] = {
            "display_name": "",
            "name": "",
        }
        nb_file_bin = base64.b64encode(
            json.dumps(nb_file_dict).encode("utf-8")
        ).decode("utf-8")

    content = [
        {
            "content": nb_file_bin,
            "content_name": os.path.basename(notebook_file),
            "contenttype_id": 1,
            "path": notebook_file if source.lower() == "lab" else None,
        }
    ]

    # Auxiliary files
    # Collect all files within the specified paths
    if aux_files[0]["source"]:
        # In the Lab, paths are relative to the user's home directory
        if source.lower() == "lab":
            # Collect all files within the specified paths
            all_aux_files = []
            for aux_file in aux_files:
                # Directories end with "/"; otherwise they are considered files; no nesting allowed
                if aux_file["source"].endswith("/"):
                    dir_cont = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": aux_file["source"]},
                    ).json()
                    dir_paths = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in dir_cont
                        if x["type"] == "file"
                    ]
                    all_aux_files.extend(dir_paths)
                else:
                    response_props = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": os.path.dirname(aux_file["source"])},
                    ).json()
                    file_path = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in response_props
                        if x["name"] == os.path.basename(aux_file["source"])
                    ]
                    all_aux_files.extend(file_path)
            # Now we have all the files with their paths and sizes in a single list
            for aux_file in all_aux_files:
                # Check file size before actually pulling it from the Lab
                assert (
                    aux_file["size"] <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                # "content" is the binary; "content_name" determines the target folder
                aux_file_bin = adaboard.request_adaboard(
                    path="jupyter/files/notebook/content/fetch",
                    params={"path": aux_file["source"]},
                ).json()["content"]
                aux_file_name = f"{aux_file['target']}{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                    "path": aux_file["source"],
                }
                content.append(tmp)

        # In the local machine, paths are absolute; nesting is allowed
        elif source.lower() == "local":
            all_aux_files = []
            for aux_file in aux_files:
                if os.path.isfile(aux_file["source"]):
                    all_aux_files.append(aux_file)
                elif os.path.isdir(aux_file["source"]):
                    for root, dirs, files in os.walk(aux_file["source"]):
                        for file in files:
                            all_aux_files.append(
                                {
                                    "source": f"{root}/{file}",
                                    "target": f"{root}/".replace(
                                        aux_file["source"], aux_file["target"]
                                    ),
                                }
                            )
            # Now we have all the files with their paths in a single list
            for aux_file in all_aux_files:
                # Check file sizes before actually opening them
                assert (
                    os.path.getsize(aux_file["source"]) <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                with open(aux_file["source"], "rb") as f:
                    aux_file_bin = base64.b64encode(f.read()).decode("utf-8")
                # "content" is the binary; "content_name" determines the target folder
                aux_file_name = f"{aux_file['target']}/{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                }
                content.append(tmp)

    acl_config = [
        {
            "acl_action": "card_view",
            "acl_type": acl_type_view if acl_type_view else "public",
            "userlist": acl_list_view if acl_type_view == "userlist" else [],
            "grouplist": acl_list_view if acl_type_view == "grouplist" else [],
        },
        {
            "acl_action": "card_launch",
            "acl_type": acl_type_launch if acl_type_launch else "public",
            "userlist": (
                acl_list_launch if acl_type_launch == "userlist" else []
            ),
            "grouplist": (
                acl_list_launch if acl_type_launch == "grouplist" else []
            ),
        },
        {
            "acl_action": "card_edit",
            "acl_type": acl_type_edit if acl_type_edit else "public",
            "userlist": acl_list_edit if acl_type_edit == "userlist" else [],
            "grouplist": acl_list_edit if acl_type_edit == "grouplist" else [],
        },
    ]
    payload = {
        "name": name,
        "description": description,
        "cardtype_id": cardtype_id,
        "author_id": adaboard.get_user()["user_id"],
        "owner_id": adaboard.get_user()["user_id"],
        "image_id": picture_id,
        "keywords": keywords_id,
        "coauthors": coauthors,
        "acls": acl_config,
        "content": content,
        "reviewers": reviewers,
    }

    response = adaboard.request_adaboard(
        path="cards",
        method=requests.post,
        json=payload,
    ).json()

    return response["id"]


def create_card_group(
    name: str,
    description: str,
    cards_id: list[int],
    acl_type_view: str = "public",
    acl_list_view: list[str] = [],
    acl_type_edit: str = "public",
    acl_list_edit: list[str] = [],
    coauthors: list[str] = [],
    keywords: list[str] = [],
    picture_id: int = 0,
    reviewers: list[str] = [],
) -> int:
    """
    Create a card group in the AdaLab Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/create_card_group.ipynb) to test this function or build upon it.

    :param name: the name of the group
    :type name: str
    :param description: description of the group (tip: supports Markdown)
    :type description: str
    :param cards_id: the IDs of the cards to be added to the group, in order of appearance
    :type cards_id: list
    :param acl_type_view: the ACL type for viewing the card group, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_view: str, optional
    :param acl_list_view: the list of users/groups allowed to view the card group, defaults to []
    :type acl_list_view: list, optional
    :param acl_type_launch: the ACL type for launching the card group, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :param acl_type_edit: the ACL type for editing the card group, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_edit: str, optional
    :param acl_list_edit: the list of users/groups allowed to edit the card group, defaults to []
    :type acl_list_edit: list, optional
    :param coauthors: the other group authors, defaults to []
    :type coauthors: list, optional
    :param keywords: the keywords related to the card group, defaults to [] (tip: must exist in the Gallery)
    :type keywords: list, optional
    :param picture_id: the ID of the picture to be used for the card group, defaults to 0
    :type picture_id: int, optional
    :param reviewers: the list of users who will review the card group, defaults to []
    :type reviewers: list, optional
    :return: the ID of the card group
    :rtype: int
    """
    validate_acl("card", False, acl_type_view, acl_type_edit)

    keywords_id = []
    if keywords:
        kws = _create_keywords_if_not_exist(keywords=keywords)
        for kw in keywords:
            keywords_id.append(kws[kw])

    acl_config = [
        {
            "acl_action": "card_view",
            "acl_type": acl_type_view if acl_type_view else "public",
            "userlist": acl_list_view if acl_type_view == "userlist" else [],
            "grouplist": acl_list_view if acl_type_view == "grouplist" else [],
        },
        {
            "acl_action": "card_launch",
            "acl_type": "logged_in",
            "userlist": [],
            "grouplist": [],
        },
        {
            "acl_action": "card_edit",
            "acl_type": acl_type_edit if acl_type_edit else "public",
            "userlist": acl_list_edit if acl_type_edit == "userlist" else [],
            "grouplist": acl_list_edit if acl_type_edit == "grouplist" else [],
        },
    ]

    group_cards = []
    for card_id in cards_id:
        group_cards.append(
            {"card_id": card_id, "order_id": cards_id.index(card_id)}
        )

    payload = {
        "name": name,
        "description": description,
        "cardtype_id": 4,
        "card_group_cards": group_cards,
        "author_id": adaboard.get_user()["user_id"],
        "owner_id": adaboard.get_user()["user_id"],
        "image_id": picture_id,
        "keywords": keywords_id,
        "coauthors": coauthors,
        "acls": acl_config,
        "reviewers": reviewers,
    }

    response = adaboard.request_adaboard(
        path="cards",
        method=requests.post,
        json=payload,
    ).json()

    return response["id"]


def create_url_card(
    name: str,
    description: str,
    remote_url: str,
    url_subtype: str,
    acl_type_view: str = "public",
    acl_list_view: list[str] = [],
    acl_type_launch: str = "public",
    acl_list_launch: list[str] = [],
    acl_type_edit: str = "public",
    acl_list_edit: list[str] = [],
    coauthors: list[str] = [],
    keywords: list[str] = [],
    picture_id: int = 0,
    reviewers: list[str] = [],
) -> int:
    """Create a URL-type card in the AdaLab Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/create_url_card.ipynb) to test this function or build upon it.

    :param name: the name of the card
    :type name: str
    :param description: the card description (tip: supports Markdown)
    :type description: str
    :param remote_url: URL to point the card to
    :type remote_url: str
    :param url_subtype: URL subtype (tip: must exist in the Gallery)
    :type url_subtype: str
    :param acl_type_view: the ACL type for viewing the card, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_view: str, optional
    :param acl_list_view: the list of users/groups allowed to view the card, defaults to []
    :type acl_list_view: list, optional
    :param acl_type_launch: the ACL type for launching the card, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_launch: str, optional
    :param acl_list_launch: the list of users/groups allowed to launch the card, defaults to []
    :type acl_list_launch: list, optional0
    :param acl_type_edit: the ACL type for editing the card, defaults to "public" (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_edit: str, optional
    :param acl_list_edit: the list of users/groups allowed to edit the card, defaults to []
    :type acl_list_edit: list, optional
    :param coauthors: the other card authors, defaults to []
    :type coauthors: list, optional
    :param keywords: the keywords related to the card, defaults to [] (tip: must exist in the Gallery)
    :type keywords: list, optional
    :param picture_id: the ID of the picture to be used for the card, defaults to 0
    :type picture_id: int, optional
    :param reviewers: the list of users who will review the card, defaults to []
    :type reviewers: list, optional
    :return: the ID of the card
    :rtype: int
    """

    validate_acl("card", False, acl_type_view, acl_type_launch, acl_type_edit)

    try:
        url_subtypes = adaboard.request_adaboard(
            "url-card-subtypes", method=requests.get
        ).json()
        url_subtype_id = [
            x["urlcardsubtype_id"]
            for x in url_subtypes
            if x["name"] == url_subtype
        ][0]
    except IndexError as e:
        raise IndexError(
            f"URL subtype '{url_subtype}' not found in the Gallery. Please use an existing URL subtype."
        ) from e

    keywords_id = []
    if keywords:
        kws = _create_keywords_if_not_exist(keywords=keywords)
        for kw in keywords:
            keywords_id.append(kws[kw])

    acl_config = [
        {
            "acl_action": "card_view",
            "acl_type": acl_type_view if acl_type_view else "public",
            "userlist": acl_list_view if acl_type_view == "userlist" else [],
            "grouplist": acl_list_view if acl_type_view == "grouplist" else [],
        },
        {
            "acl_action": "card_launch",
            "acl_type": acl_type_launch if acl_type_launch else "public",
            "userlist": (
                acl_list_launch if acl_type_launch == "userlist" else []
            ),
            "grouplist": (
                acl_list_launch if acl_type_launch == "grouplist" else []
            ),
        },
        {
            "acl_action": "card_edit",
            "acl_type": acl_type_edit if acl_type_edit else "public",
            "userlist": acl_list_edit if acl_type_edit == "userlist" else [],
            "grouplist": acl_list_edit if acl_type_edit == "grouplist" else [],
        },
    ]

    payload = {
        "name": name,
        "description": description,
        "cardtype_id": 3,
        "remote_url": remote_url,
        "url_subtype_id": url_subtype_id,
        "author_id": adaboard.get_user()["user_id"],
        "owner_id": adaboard.get_user()["user_id"],
        "image_id": picture_id,
        "keywords": keywords_id,
        "coauthors": coauthors,
        "acls": acl_config,
        "reviewers": reviewers,
    }

    response = adaboard.request_adaboard(
        "cards", json=payload, method=requests.post
    ).json()

    return response["id"]


def create_url_subtype(
    name: str = "",
    description: str = "",
    picture_id: int = 0,
    protected: bool = False,
) -> int:
    """
    Create a new URL subtype in the AdaLab Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/cards/create_url_subtype.ipynb) to test this function or build upon it.

    :param name: the name of the URL subtype
    :type name: str, optional
    :param description: the description of the URL subtype
    :type description: str, optional
    :param picture_id: the ID of the picture to be used for the card
    :type picture_id: int, optional
    :param protected: whether to protect the URL subtype from deletion, defaults to False
    :type protected: bool, optional
    :return: the ID of the URL subtype
    :rtype: int
    """

    payload = {
        "name": name,
        "description": description,
        "image_id": picture_id,
        "protected": protected,
    }
    response = adaboard.request_adaboard(
        "url-card-subtypes", json=payload, method=requests.post
    ).json()

    return response["id"]


def delete_card(card_id: int) -> None | str:
    """
    Delete a card from the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/delete_card.ipynb) to test this function or build upon it.

    :param card_id: the id of the card to be deleted
    :type card_id: int
    :return: Nothing
    :rtype: None
    """

    adaboard.request_adaboard(
        path=f"cards/{card_id}",
        method=requests.delete,
    )

    return None


def delete_card_group(group_id: int) -> None:
    """
    Delete a card group from the AdaLab Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/delete_card_group.ipynb) to test this function or build upon it.

    :param group_id: the ID of the group to be deleted
    :type group_id: int
    """

    adaboard.request_adaboard(
        path=f"cards/{group_id}",
        method=requests.delete,
    )

    return None


def delete_url_subtype(url_subtype_id: int) -> None:
    """
    Delete a URL subtype from the AdaLab Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/cards/delete_url_subtype.ipynb) to test this function or build upon it.

    :param url_subtype_id: the ID of the URL subtype
    :type url_subtype_id: int
    :return: Nothing
    :rtype: None
    """

    adaboard.request_adaboard(
        path=f"url-card-subtypes/{url_subtype_id}",
        method=requests.delete,
    )

    return None


def edit_card(  # noqa: C901
    card_id: int,
    name: str = None,
    description: str = None,
    cardtype: str = None,
    notebook_file: str = None,
    acl_type_view: str = None,
    acl_list_view: list[str] = None,
    acl_type_launch: str = None,
    acl_list_launch: list[str] = None,
    acl_type_edit: str = None,
    acl_list_edit: list[str] = None,
    aux_files: dict[str, str] = None,
    coauthors: list[str] = None,
    keep_aux_files: bool = True,
    keywords: list[str] = None,
    picture_id: int = None,
    reviewers: list[str] = None,
    source: str = "lab",
) -> None:
    """
    Edit a card in the AdaLab Gallery. Note that unspecified fields will not be changed.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/edit_card.ipynb) to test this function or build upon it.

    :param card_id: the ID of the card
    :type card_id: int
    :param name: the new name of the card
    :type name: str, optional
    :param description: the new description of the card (tip: supports Markdown)
    :type description: str, optional
    :param cardtype: the type of card (tip: must be "notebook" or "voila")
    :type cardtype: str, optional
    :param notebook_file: the path to the new main notebook file to be added to the card
    :type notebook_file: str, optional
    :param acl_type_view: the ACL type for viewing the card (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_view: str, optional
    :param acl_list_view: the list of users/groups allowed to view the card
    :type acl_list_view: list, optional
    :param acl_type_launch: the ACL type for launching the card (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_launch: str, optional
    :param acl_list_launch: the list of users/groups allowed to launch the card
    :type acl_list_launch: list, optional
    :param acl_type_edit: the ACL type for editing the card (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_edit: str, optional
    :param acl_list_edit: the list of users/groups allowed to edit the card
    :type acl_list_edit: list, optional
    :param aux_files: the list of paths to the new auxiliary files to be added to the card
    :type aux_files: dict, optional
    :param coauthors: the other card authors
    :type coauthors: list, optional
    :param keep_aux_files: whether to keep the existing auxiliary files or not, defaults to True
    :type keep_aux_files: bool, optional
    :param keywords: the keywords related to the card, defaults to [] (tip: must exist in the Gallery)
    :type keywords: list, optional
    :param picture_id: the ID of the picture to be used for the card
    :type picture_id: int, optional
    :param reviewers: the list of users who will review the card
    :type reviewers: list, optional
    :param source: the source of the card files, defaults to "lab" (tip: must be "lab" or "local")
    :type source: str, optional
    :return: nothing
    :rtype: None
    """
    assert (
        cardtype is None or cardtype.lower() in ALLOWED_CARD_TYPES
    ), "Card type must be 'notebook' or 'voila'."
    assert notebook_file is None or notebook_file.lower().endswith(
        ".ipynb"
    ), "Notebook file must be a .ipynb file."
    assert (
        source.lower() in ALLOWED_CARD_SOURCES
    ), "Files source must be 'lab' or 'local'."

    validate_acl("card", True, acl_type_view, acl_type_launch, acl_type_edit)
    assert aux_files is None or all(
        issubclass(type(x), dict) for x in aux_files
    ), "Auxiliary files must be a list of dictionaries."
    assert aux_files is None or all(
        ("source" in d and "target" in d) for d in aux_files
    ), "Auxiliary files must have keys 'source' and 'target'."

    try:
        cardtype_id = ALLOWED_CARD_TYPES[cardtype.lower()]
    except AttributeError:
        cardtype_id = None

    keywords_id = []
    if keywords:
        kws = _create_keywords_if_not_exist(keywords=keywords)
        for kw in keywords:
            keywords_id.append(kws[kw])

    # Collect new configuration options
    card_config = {
        "name": name,
        "description": description,
        "cardtype_id": cardtype_id,
        "author_id": None,
        "owner_id": None,
        "image_id": picture_id,
        "keywords": keywords_id,
        "coauthors": coauthors,
        "acls": None,
        "content": None,
        "reviewers": reviewers,
    }
    # Fetch old configuration
    old_config = adaboard.request_adaboard(
        path=f"cards/{card_id}", method=requests.get
    ).json()

    # ACLs need to be re-formatted for the PUT request
    old_acls = []
    for tmp in old_config["acls"]:
        if tmp["acl_type"] == "grouplist":
            tmp["grouplist"] = [x["group_id"] for x in tmp["grouplist"]]
        elif tmp["acl_type"] == "userlist":
            tmp["userlist"] = [x["user_id"] for x in tmp["userlist"]]
        else:
            tmp["grouplist"] = []
            tmp["userlist"] = []
        old_acls.append(tmp)
    old_config["acls"] = old_acls
    # Collect the new ACLs
    acl_config = [
        {
            "acl_action": "card_view",
            "acl_type": (
                acl_type_view
                if acl_type_view
                else old_config["acls"][0]["acl_type"]
            ),
            "userlist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["userlist"]
            ),
            "grouplist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["grouplist"]
            ),
        },
        {
            "acl_action": "card_launch",
            "acl_type": (
                acl_type_launch
                if acl_type_launch
                else old_config["acls"][1]["acl_type"]
            ),
            "userlist": (
                acl_list_launch
                if acl_type_launch
                else old_config["acls"][1]["userlist"]
            ),
            "grouplist": (
                acl_list_launch
                if acl_type_launch
                else old_config["acls"][1]["grouplist"]
            ),
        },
        {
            "acl_action": "card_edit",
            "acl_type": (
                acl_type_edit
                if acl_type_edit
                else old_config["acls"][2]["acl_type"]
            ),
            "userlist": (
                acl_list_edit
                if acl_type_edit
                else old_config["acls"][2]["userlist"]
            ),
            "grouplist": (
                acl_list_edit
                if acl_type_edit
                else old_config["acls"][2]["grouplist"]
            ),
        },
    ]
    card_config["acls"] = acl_config
    # Author, coauthors and reviewers need to be re-formatted for the PUT request
    old_author = old_config["author"]["user_id"]
    old_config["author"] = old_author
    old_coauthors = [
        coauthor["user_id"] for coauthor in old_config["coauthors"]
    ]
    old_config["coauthors"] = old_coauthors
    old_reviewers = [
        reviewer["user_id"] for reviewer in old_config["reviewers"]
    ]
    old_config["reviewers"] = old_reviewers
    # Update content
    # First, the main Notebook
    if notebook_file:
        if source.lower() == "lab":
            # Check file size before actually pulling it from the Lab
            try:
                response_props = adaboard.request_adaboard(
                    path="jupyter/files/notebook/content/",
                    params={"path": os.path.dirname(notebook_file)},
                ).json()
                props = [
                    x
                    for x in response_props
                    if x["name"] == os.path.basename(notebook_file)
                ][0]
            except IndexError as e:
                raise IndexError(
                    f"Notebook file '{notebook_file}' not found in the Lab."
                ) from e
            assert (
                props["size"] <= MAX_FILE_SIZE
            ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
            nb_file_bin = adaboard.request_adaboard(
                path="jupyter/files/notebook/content/fetch",
                params={"path": notebook_file},
            ).json()["content"]
        else:
            # Check file size before actually opening it
            assert (
                os.path.getsize(notebook_file) <= MAX_FILE_SIZE
            ), f"Notebook file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
            with open(notebook_file, "rb") as f:
                nb_file_bin = base64.b64encode(f.read()).decode("utf-8")

        content = [
            {
                "content": nb_file_bin,
                "content_name": os.path.basename(notebook_file),
                "contenttype_id": 1,
                "path": notebook_file if source.lower() == "lab" else None,
            }
        ]
    else:
        content = [
            x for x in old_config["content"] if x["contenttype_id"] == 1
        ]

    # Next, the auxiliary files
    # Start by collecting or ditching the old ones
    if keep_aux_files:
        old_aux_files = [
            x for x in old_config["content"] if x["contenttype_id"] == 2
        ]
        content.extend(old_aux_files)
    # Then, add the new ones
    if aux_files:
        # In the Lab, paths are relative to the user's home directory
        if source.lower() == "lab":
            # Collect all files within the specified paths
            all_aux_files = []
            for aux_file in aux_files:
                # Directories end with "/"; otherwise they are considered files; no nesting allowed
                if aux_file["source"].endswith("/"):
                    dir_cont = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": aux_file["source"]},
                    ).json()
                    dir_paths = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in dir_cont
                        if x["type"] == "file"
                    ]
                    all_aux_files.extend(dir_paths)
                else:
                    response_props = adaboard.request_adaboard(
                        path="jupyter/files/notebook/content/",
                        params={"path": os.path.dirname(aux_file["source"])},
                    ).json()
                    file_path = [
                        {
                            "source": x["path"],
                            "target": aux_file["target"],
                            "size": x["size"],
                        }
                        for x in response_props
                        if x["name"] == os.path.basename(aux_file["source"])
                    ]
                    all_aux_files.extend(file_path)
            # Now we have all the files with their paths and sizes in a single list
            for aux_file in all_aux_files:
                # Check file size before actually pulling it from the Lab
                assert (
                    aux_file["size"] <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                # "content" is the binary; "content_name" determines the target folder
                aux_file_bin = adaboard.request_adaboard(
                    path="jupyter/files/notebook/content/fetch",
                    params={"path": aux_file["source"]},
                ).json()["content"]
                aux_file_name = f"{aux_file['target']}{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                    "path": aux_file["source"],
                }
                content.append(tmp)

        # In the local machine, paths are absolute; nesting is allowed
        elif source.lower() == "local":
            all_aux_files = []
            for aux_file in aux_files:
                if os.path.isfile(aux_file["source"]):
                    all_aux_files.append(aux_file)
                elif os.path.isdir(aux_file["source"]):
                    for root, dirs, files in os.walk(aux_file["source"]):
                        for file in files:
                            all_aux_files.append(
                                {
                                    "source": f"{root}/{file}",
                                    "target": f"{root}/".replace(
                                        aux_file["source"], aux_file["target"]
                                    ),
                                }
                            )
            # Now we have all the files with their paths in a single list
            for aux_file in all_aux_files:
                # Check file sizes before actually opening them
                assert (
                    os.path.getsize(aux_file["source"]) <= MAX_FILE_SIZE
                ), f"Auxiliary file size exceeds the maximum allowed ({MAX_FILE_SIZE} bytes). Please reduce the size of the file and try again."
                with open(aux_file["source"], "rb") as f:
                    aux_file_bin = base64.b64encode(f.read()).decode("utf-8")
                # "content" is the binary; "content_name" determines the target folder
                aux_file_name = f"{aux_file['target']}/{os.path.basename(aux_file['source'])}"
                tmp = {
                    "content": aux_file_bin,
                    "content_name": aux_file_name,
                    "contenttype_id": 2,
                }
                content.append(tmp)

    card_config["content"] = content
    # Build request payload combining new and old options
    payload = {
        k: v if v is not None else old_config[k]
        for k, v in card_config.items()
    }
    adaboard.request_adaboard(
        f"cards/{card_id}", method=requests.put, json=payload
    )
    return None


def edit_card_group(
    group_id: int,
    name: str = None,
    description: str = None,
    cards_id: list[int] = None,
    acl_type_view: str = None,
    acl_list_view: list[str] = None,
    acl_type_edit: str = None,
    acl_list_edit: list[str] = None,
    coauthors: list[str] = None,
    keywords: list[str] = None,
    picture_id: int = None,
    reviewers: list[str] = None,
) -> None:
    """
    Edit a card group in the AdaLab Gallery. Note that the unspecified fields will not be modified.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/edit_card_group.ipynb) to test this function or build upon it.

    :param group_id: the ID of the card group
    :type group_id: int
    :param name: the new name of the card group
    :type name: str, optional
    :param description: the new description of the card group
    :type description: str, optional
    :param cards_id: the IDs of the cards to be added to the group
    :type cards_id: list, optional
    :param acl_type_view: the ACL type for viewing the card group (tip: must be "public", "logged_in", "userlist" or "grouplist"), defaults to None
    :type acl_type_view: str, optional
    :param acl_list_view: the list of users/groups allowed to view the card group
    :type acl_list_view: list, optional
    :param acl_type_edit: the ACL type for editing the card group (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_edit: str, optional
    :param acl_list_edit: the list of users/groups allowed to edit the card group
    :type acl_list_edit: list, optional
    :param coauthors: the other authors of the card group
    :type coauthors: list, optional
    :param keywords: the keywords related to the card group
    :type keywords: list, optional
    :param picture_id: the ID of the new picture to be used for the card group
    :type picture_id: int, optional
    :param reviewers: the list of users who will review the card group
    :type reviewers: list, optional
    """

    validate_acl("card", True, acl_type_view, acl_type_edit)

    keywords_id = []
    if keywords:
        kws = _create_keywords_if_not_exist(keywords=keywords)
        for kw in keywords:
            keywords_id.append(kws[kw])

    # Collect new configuration options
    card_group_config = {
        "name": name,
        "description": description,
        "cardtype_id": 4,
        "card_group_cards": None,
        "author_id": None,
        "owner_id": None,
        "image_id": picture_id,
        "keywords": keywords_id,
        "coauthors": coauthors,
        "acls": None,
        "reviewers": reviewers,
    }
    # Fetch old configuration
    old_config = adaboard.request_adaboard(
        path=f"cards/{group_id}", method=requests.get
    ).json()

    # ACLs need to be re-formatted for the PUT request
    old_acls = []
    for tmp in old_config["acls"]:
        if tmp["acl_type"] == "grouplist":
            tmp["grouplist"] = [x["group_id"] for x in tmp["grouplist"]]
        elif tmp["acl_type"] == "userlist":
            tmp["userlist"] = [x["user_id"] for x in tmp["userlist"]]
        else:
            tmp["grouplist"] = []
            tmp["userlist"] = []
        old_acls.append(tmp)
    old_config["acls"] = old_acls
    # Collect the new ACLs
    acl_config = [
        {
            "acl_action": "card_view",
            "acl_type": (
                acl_type_view
                if acl_type_view
                else old_config["acls"][0]["acl_type"]
            ),
            "userlist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["userlist"]
            ),
            "grouplist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["grouplist"]
            ),
        },
        {
            "acl_action": "card_launch",
            "acl_type": "logged_in",
            "userlist": [],
            "grouplist": [],
        },
        {
            "acl_action": "card_edit",
            "acl_type": (
                acl_type_edit
                if acl_type_edit
                else old_config["acls"][2]["acl_type"]
            ),
            "userlist": (
                acl_list_edit
                if acl_type_edit
                else old_config["acls"][2]["userlist"]
            ),
            "grouplist": (
                acl_list_edit
                if acl_type_edit
                else old_config["acls"][2]["grouplist"]
            ),
        },
    ]
    card_group_config["acls"] = acl_config
    # Author, coauthors and reviewers need to be re-formatted for the PUT request
    old_author = old_config["author"]["user_id"]
    old_config["author"] = old_author
    old_coauthors = [
        coauthor["user_id"] for coauthor in old_config["coauthors"]
    ]
    old_config["coauthors"] = old_coauthors
    old_reviewers = [
        reviewer["user_id"] for reviewer in old_config["reviewers"]
    ]
    old_config["reviewers"] = old_reviewers

    # Update the card group cards
    if cards_id:
        old_config["card_group_cards"] = []
        for card_id in cards_id:
            old_config["card_group_cards"].append(
                {"card_id": card_id, "order_id": cards_id.index(card_id)}
            )

    # Build request payload combining new and old options
    payload = {
        k: v if v is not None else old_config[k]
        for k, v in card_group_config.items()
    }
    adaboard.request_adaboard(
        path=f"cards/{group_id}",
        method=requests.put,
        json=payload,
    )

    return None


def edit_url_card(
    card_id: int,
    name: str = None,
    description: str = None,
    remote_url: str = None,
    url_subtype: str = None,
    acl_type_view: str = None,
    acl_list_view: list[str] = None,
    acl_type_launch: str = None,
    acl_list_launch: list[str] = None,
    acl_type_edit: str = None,
    acl_list_edit: list[str] = None,
    coauthors: list[str] = None,
    keywords: list[str] = None,
    picture_id: int = None,
    reviewers: list[str] = None,
) -> None:
    """
    Edit a card of the URL type in the AdaLab Gallery. Note that unspecified fields will not be changed.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/edit_url_card.ipynb) to test this function or build upon it.

    :param card_id: the ID of the card
    :type card_id: int
    :param name: the new name of the card
    :type name: str, optional
    :param description: the new description of the card (tip: supports Markdown)
    :type description: str, optional
    :param remote_url: _description_, defaults to None
    :type remote_url: str, optional
    :param url_subtype: _description_, defaults to None
    :type url_subtype: str, optional
    :param acl_type_view: the ACL type for viewing the card (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_view: str, optional
    :param acl_list_view: the list of users/groups allowed to view the card
    :type acl_list_view: list, optional
    :param acl_type_launch: the ACL type for launching the card (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_launch: str, optional
    :param acl_list_launch: the list of users/groups allowed to launch the card
    :type acl_list_launch: list, optional
    :param acl_type_edit: the ACL type for editing the card (tip: must be "public", "logged_in", "userlist" or "grouplist")
    :type acl_type_edit: str, optional
    :param acl_list_edit: the list of users/groups allowed to edit the card
    :type acl_list_edit: list, optional
    :param coauthors: _description_, defaults to None
    :param coauthors: the other card authors
    :type coauthors: list, optional
    :param keywords: the keywords related to the card, defaults to [] (tip: must exist in the Gallery)
    :type keywords: list, optional
    :param picture_id: the ID of the picture to be used for the card
    :type picture_id: int, optional
    :param reviewers: the list of users who will review the card
    :type reviewers: list, optional
    :return: nothing
    :rtype: None
    """

    validate_acl("card", True, acl_type_view, acl_type_launch, acl_type_edit)

    try:
        url_subtypes = adaboard.request_adaboard(
            "url-card-subtypes", method=requests.get
        ).json()
        url_subtype_id = [
            x["urlcardsubtype_id"]
            for x in url_subtypes
            if x["name"] == url_subtype
        ][0]
    except IndexError as e:
        raise IndexError(
            f"URL subtype '{url_subtype}' not found in the Gallery. Please use an existing URL subtype."
        ) from e

    keywords_id = []
    if keywords:
        kws = _create_keywords_if_not_exist(keywords=keywords)
        for kw in keywords:
            keywords_id.append(kws[kw])

    # Collect new configuration options
    card_config = {
        "name": name,
        "description": description,
        "cardtype_id": 3,
        "remote_url": remote_url,
        "url_subtype_id": url_subtype_id,
        "author_id": None,
        "owner_id": None,
        "image_id": picture_id,
        "keywords": keywords_id,
        "coauthors": coauthors,
        "acls": None,
        "reviewers": reviewers,
    }
    # Fetch old configuration
    old_config = adaboard.request_adaboard(
        path=f"cards/{card_id}", method=requests.get
    ).json()

    # ACLs need to be re-formatted for the PUT request
    old_acls = []
    for tmp in old_config["acls"]:
        if tmp["acl_type"] == "grouplist":
            tmp["grouplist"] = [x["group_id"] for x in tmp["grouplist"]]
        elif tmp["acl_type"] == "userlist":
            tmp["userlist"] = [x["user_id"] for x in tmp["userlist"]]
        else:
            tmp["grouplist"] = []
            tmp["userlist"] = []
        old_acls.append(tmp)
    old_config["acls"] = old_acls
    # Collect the new ACLs
    acl_config = [
        {
            "acl_action": "card_view",
            "acl_type": (
                acl_type_view
                if acl_type_view
                else old_config["acls"][0]["acl_type"]
            ),
            "userlist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["userlist"]
            ),
            "grouplist": (
                acl_list_view
                if acl_list_view
                else old_config["acls"][0]["grouplist"]
            ),
        },
        {
            "acl_action": "card_launch",
            "acl_type": (
                acl_type_launch
                if acl_type_launch
                else old_config["acls"][1]["acl_type"]
            ),
            "userlist": (
                acl_list_launch
                if acl_type_launch
                else old_config["acls"][1]["userlist"]
            ),
            "grouplist": (
                acl_list_launch
                if acl_type_launch
                else old_config["acls"][1]["grouplist"]
            ),
        },
        {
            "acl_action": "card_edit",
            "acl_type": (
                acl_type_edit
                if acl_type_edit
                else old_config["acls"][2]["acl_type"]
            ),
            "userlist": (
                acl_list_edit
                if acl_type_edit
                else old_config["acls"][2]["userlist"]
            ),
            "grouplist": (
                acl_list_edit
                if acl_type_edit
                else old_config["acls"][2]["grouplist"]
            ),
        },
    ]
    card_config["acls"] = acl_config
    # Author, coauthors and reviewers need to be re-formatted for the PUT request
    old_author = old_config["author"]["user_id"]
    old_config["author"] = old_author
    old_coauthors = [
        coauthor["user_id"] for coauthor in old_config["coauthors"]
    ]
    old_config["coauthors"] = old_coauthors
    old_reviewers = [
        reviewer["user_id"] for reviewer in old_config["reviewers"]
    ]
    old_config["reviewers"] = old_reviewers

    # Build request payload combining new and old options
    payload = {
        k: v if v is not None else old_config[k]
        for k, v in card_config.items()
    }
    adaboard.request_adaboard(
        f"cards/{card_id}", method=requests.put, json=payload
    )
    return None


def edit_url_subtype(
    url_subtype_id: int,
    name: str = None,
    description: str = None,
    picture_id: int = None,
    protected: bool = None,
) -> None:
    """
    Edit a URL subtype in the AdaLab Gallery. Note that unspecified fields will not be changed.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/admin/cards/edit_url_subtype.ipynb) to test this function or build upon it.

    :param url_subtype_id: the ID of the URL subtype
    :type url_subtype_id: int
    :param name: the new name of the URL subtype
    :type name: str, optional
    :param description: the new description of the URL subtype
    :type description: str, optional
    :param picture_id: the ID of the picture to be used for the URL subtype
    :type picture_id: int, optional
    :param protected: whether to protect the URL subtype from deletion, defaults to False
    :type protected: bool, optional
    :return: nothing
    :rtype: None
    """

    # Collect new configuration options
    url_subtype_config = {
        "name": name,
        "description": description,
        "image_id": picture_id,
        "protected": protected,
    }
    # Fetch old configuration
    tmp = adaboard.request_adaboard(
        path="url-card-subtypes", method=requests.get
    ).json()

    old_config = [x for x in tmp if x["urlcardsubtype_id"] == url_subtype_id][
        0
    ]

    # Build request payload combining new and old options
    payload = {
        k: v if v is not None else old_config[k]
        for k, v in url_subtype_config.items()
    }
    adaboard.request_adaboard(
        f"url-card-subtypes/{url_subtype_id}",
        method=requests.put,
        json=payload,
    )
    return None


def expose_card(card_id: int) -> None:
    """
    Expose (un-hide) a specific card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/expose_card.ipynb) to test this function or build upon it.

    :param card_id: the id of the card to be exposed
    :type card_id: int
    :return: Nothing
    :rtype: None
    """

    set_card_visibility(
        card_id=card_id,
        new_status="exposed",
    )

    return None


def get_cards(card_type: str = "all") -> list[list[int | str]]:
    """
    Gets all cards registered in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/get_cards.ipynb) to test this function or build upon it.

    :param card_type: The type of cards to return, defaults to "all".
    Available values are "all", "notebook", "voila", "url" and "kernel".
    :type card_type: str
    :return: a list with the main info of all cards in the Gallery
    :rtype: list
    """
    card_types = [
        x.get("key")
        for x in adaboard.request_adaboard(path="cardtypes").json()
    ]
    card_types.append("all")
    assert (
        card_type in card_types
    ), f"Card type must be one of {', '.join(card_types)}"
    all_cards = adaboard.get_all_pages(api_url="cards")

    if card_type == "all":
        return [
            [
                x.get("card_id"),
                x.get("cardtype").get("key"),
                x.get("name"),
                x.get("author").get("name"),
            ]
            for x in all_cards
        ]
    return [
        [
            x.get("card_id"),
            x.get("cardtype").get("key"),
            x.get("name"),
            x.get("author").get("name"),
        ]
        for x in all_cards
        if x.get("cardtype").get("key") == card_type
    ]


def get_card_contents(card_id: int) -> dict[str, Union[str, list[dict]]]:
    """
    Gets the contents of a specific card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/get_card_contents.ipynb) to test this function or build upon it.

    :param card_id: the id of the card to be fetched
    :type card_id: int
    :return: a dictionary with the different fields in the card
    :rtype: dict
    """

    response = adaboard.request_adaboard(path=f"cards/{card_id}").json()

    return_struct = {
        "card_author": response.get("author").get("name"),
        "card_name": response.get("name"),
        "description": response.get("description"),
        "card_type": response.get("cardtype").get("key"),
        "contents": [],
    }

    for contents in response.get("content"):
        content = base64.b64decode(contents.get("content"))
        contents["content"] = content
        return_struct.get("contents").append(contents)

    return return_struct


def get_card_issues(
    card_id: int,
) -> list[dict[str, int | str | dict[str, str | int]]]:
    """
    Get the issues related to a specific card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/get_card_issues.ipynb) to test this function or build upon it.

    :param card_id: the id of the card whose issues are to be fetched
    :type card_id: int
    :return: list of dictionaries with issues content and metadata
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path=f"cards/{card_id}/issues", method=requests.get
    ).json()

    return response


def get_card_types() -> list[dict[str, Union[int, str, bool]]]:
    """
    List all the card types in the Gallery and their main information.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/get_card_types.ipynb) to test this function or build upon it.

    :return: list of card-types dictionaries
    :rtype: list
    """

    response = adaboard.request_adaboard(
        path="cardtypes",
        method=requests.get,
    ).json()

    return response


def get_card_types_stats() -> list[dict[str, str | int]]:
    """
    Get the statistics of the card types in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/get_card_types_stats.ipynb) to test this function or build upon it.

    :return: list of card-types-info dictionaries
    :rtype: list
    """
    response = adaboard.request_adaboard(
        path="cardtypes/statistics", method=requests.get
    ).json()

    return response


def get_url_subtypes() -> list[dict[str, str | int | bool]]:
    """
    Get the URL subtypes available in the AdaLab Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/get_url_subtypes.ipynb) to test this function or build upon it.

    :return: list with information about existing URL subtypes
    :rtype: list[dict[str, str | int | bool]]
    """

    response = adaboard.request_adaboard(
        path="url-card-subtypes",
        method=requests.get,
    ).json()

    return response


def hide_card(card_id: int) -> None:
    """
    Hide a specific card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/hide_card.ipynb) to test this function or build upon it.

    :param card_id: the id of the card to be hidden
    :type card_id: int
    :return: Nothing
    :rtype: None
    """

    set_card_visibility(card_id=card_id, new_status="hidden")

    return None


def launch_card(
    card_id: int, start_lab: bool = True, lab_config: dict = {}
) -> str:
    """
    Launch a card from the Gallery into the user's Lab.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/launch_card.ipynb) to test this function or build upon it.

    :param card_id: the ID of the card to be launched
    :type card_id: int
    :param start_lab: whether to also start the Lab when launching the card, defaults to True. Note that the Lab must be stopped for this to have any effect
    :type start_lab: bool, optional
    :param lab_config: desired configuration for the Lab environment. It can be set up either passing the IDs of the values for each configuration type, or by choosing one of the user's Lab profiles. Defaults to None (i.e., uses the user's default Lab profile). Can
    :type lab_config: dict, optional
    :return: URL/path to Notebook file in the user's Lab
    :rtype: str
    """

    # Check whether the chosen configuration is valid
    if lab_config:
        assert list(lab_config.keys()) == ["options"] or list(
            lab_config.keys()
        ) == [
            "profile_id"
        ], "Configuration must be a list of IDs with key 'options', or a value with key 'profile_id'."
        if list(lab_config.keys()) == ["options"]:
            config_opts = adaboard.request_adaboard(
                path="jupyter/system/options"
            ).json()
            tmp = []
            for item in [x["configuration_type_options"] for x in config_opts]:
                tmp.extend(item)
            config_opts_ids = [x["id"] for x in tmp]
            for id in lab_config["options"]:
                assert (
                    id in config_opts_ids
                ), "Invalid configuration option ID."
        elif list(lab_config.keys()) == ["profile_id"]:
            config_profiles = adaboard.request_adaboard(
                path="jupyter/system/configurationprofiles"
            ).json()
            config_profiles_ids = [x["id"] for x in config_profiles]
            assert (
                lab_config["profile_id"] in config_profiles_ids
            ), "Invalid configuration profile ID."

    payload = {
        "start_lab": start_lab,
        "lab_configuration": lab_config,
    }
    response = adaboard.request_adaboard(
        path=f"cards/{card_id}/launch", method=requests.post, json=payload
    ).json()

    card_path = response["message"]

    # Set up the card kernel
    # Get the main Notebook
    nb_path = card_path[card_path.find("/gallery") + 1 :]
    response = adaboard.request_adaboard(
        path="jupyter/files/notebook/content/fetch", params={"path": nb_path}
    ).json()
    nb_file_dict = json.loads(
        base64.b64decode(response["content"]).decode("utf-8")
    )

    # Check if the kernel is custom
    kernel_pattern = re.compile(r"^(kp|kt)(\d+)$")
    kernel_match = kernel_pattern.match(
        nb_file_dict["metadata"]["kernelspec"]["name"]
    )
    if kernel_match:
        # Check if the kernel linked to the card is installed. If not, install it
        id_dict = {}
        kernels = adaboard.request_adaboard(path="jupyter/kernelspecs").json()
        if kernel_match.group(1) == "kp":
            kernel_metadata_id = int(kernel_match.group(2))
            try:
                [x for x in kernels if x["metadata_id"] == kernel_metadata_id][
                    0
                ]
            except IndexError:
                id_dict = {"metadata_id": kernel_metadata_id}
        elif kernel_match.group(1) == "kt":
            kernel_run_id = int(kernel_match.group(2))
            try:
                [x for x in kernels if x["run_id"] == kernel_run_id][0]
            except IndexError:
                id_dict = {"run_id": kernel_run_id}
        # If the kernel is not installed and it exists, install it
        if id_dict:
            payload = {
                "display_name": nb_file_dict["metadata"]["kernelspec"][
                    "display_name"
                ],
                "include_dummy_notebook": False,
                "start_options": {"start_lab": False, "lab_configuration": {}},
            }
            payload.update(id_dict)
            adaboard.request_adaboard(
                path="jupyter/kernelspecs",
                method=requests.post,
                json=payload,
            )

    return card_path


def set_card_visibility(card_id: int, new_status: str) -> None:
    """
    Set the visibility status of a specific card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/set_card_visibility.ipynb) to test this function or build upon it.

    :param card_id: the id of the card whose visibility is to be set
    :type card_id: int
    :param new_status: visibility to be set
    :type new_status: str
    :return: Nothing
    :rtype: None
    """

    assert new_status.lower() in [
        "exposed",
        "hidden",
    ], "Status must be 'exposed' or 'hidden'"

    status_dict = {"exposed": "false", "hidden": "true"}
    new_status = new_status.lower()
    payload = {"hidden": status_dict[new_status]}
    adaboard.request_adaboard(
        path=f"cards/{card_id}/hidden", method=requests.put, json=payload
    )

    return None


def toggle_card_favorite(card_id: int) -> str:
    """
    Toggle the favorite status of a card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/toggle_card_favorite.ipynb) to test this function or build upon it.

    :param card_id: the id of the card whose status is to be toggled
    :type card_id: int
    :return: a message describing new status
    :rtype: str
    """

    response = adaboard.request_adaboard(
        path=f"favorite/{card_id}", method=requests.post
    ).json()

    return response["message"]


def toggle_card_review(card_id: int) -> None:
    """
    Toggle the review status of a card in the Gallery.

    Use the [example Jupyter Notebook](https://github.com/adamatics/adalib_example_notebooks/blob/main/user/cards/toggle_card_review.ipynb) to test this function or build upon it.

    :param card_id: the id of the card whose status is to be toggled
    :type card_id: int
    :return: nothing
    :rtype: None
    """

    adaboard.request_adaboard(
        path=f"cards/{card_id}/review", method=requests.put
    )

    return None



"""General module for custom exceptions.
"""


class QueryLimitReached(ValueError):
    """Exception to signal that the limit of queries in Superset has been
    reached.

    :param ValueError: The ValueError of the exception
    :type ValueError: ValueError
    """

    pass



"""The adalib base module exposes the core integrations of all the components in the AdaLab platform.
"""

import importlib.metadata

_DISTRIBUTION_METADATA = importlib.metadata.metadata("adalib")
__project__ = _DISTRIBUTION_METADATA["name"]
__version__ = _DISTRIBUTION_METADATA["version"]
__description__ = _DISTRIBUTION_METADATA["description"]

__all__ = [
    "adaboard",
    "apps",
    "cards",
    "harbor",
    "issues",
    "keywords",
    "lab",
    "schedules",
    "superset",
]
